"""AFL AFLTables player model."""

# pylint: disable=line-too-long,duplicate-code,too-many-arguments,too-many-locals
import datetime
import logging
import os
from urllib.parse import urlparse

import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from ....cache import MEMORY
from ...player_model import VERSION, PlayerModel
from ...sex import Sex
from ...species import Species


def _create_afl_afltables_player_model(
    player_url: str,
    jersey: str,
    kicks: int | None,
    name: str,
    marks: int | None,
    handballs: int | None,
    disposals: int | None,
    goals: int | None,
    behinds: int | None,
    hit_outs: int | None,
    tackles: int | None,
    rebounds: int | None,
    insides: int | None,
    clearances: int | None,
    clangers: int | None,
    free_kicks_for: int | None,
    free_kicks_against: int | None,
    brownlow_votes: int | None,
    contested_possessions: int | None,
    uncontested_possessions: int | None,
    contested_marks: int | None,
    marks_inside: int | None,
    one_percenters: int | None,
    bounces: int | None,
    goal_assists: int | None,
    percentage_played: float | None,
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    version: str,
) -> PlayerModel:
    o = urlparse(player_url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)
    jersey = "".join(filter(str.isdigit, jersey))
    response = session.get(player_url)
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "lxml")
    player_page_text = soup.get_text()

    birth_date = None
    try:
        birth_date = parse(
            player_page_text.split("Born:")[1].strip().split()[0].strip()
        )
    except IndexError:
        logging.debug("Couldn't find birth date from %s", response.url)

    weight = None
    try:
        weight = float(player_page_text.split("Weight:")[1].strip().split()[0].strip())
    except IndexError:
        logging.debug("Couldn't find weight from %s", response.url)

    height = None
    try:
        height = float(player_page_text.split("Height:")[1].strip().split()[0].strip())
    except IndexError:
        logging.debug("Couldn't find height from %s", response.url)

    return PlayerModel(
        identifier=identifier,
        jersey=jersey,
        kicks=kicks,
        fumbles=None,
        fumbles_lost=None,
        field_goals=None,
        field_goals_attempted=None,
        offensive_rebounds=None,
        assists=None,
        turnovers=None,
        name=name,
        marks=marks,
        handballs=handballs,
        disposals=disposals,
        goals=goals,
        behinds=behinds,
        hit_outs=hit_outs,
        tackles=tackles,
        rebounds=rebounds,
        insides=insides,
        clearances=clearances,
        clangers=clangers,
        free_kicks_for=free_kicks_for,
        free_kicks_against=free_kicks_against,
        brownlow_votes=brownlow_votes,
        contested_possessions=contested_possessions,
        uncontested_possessions=uncontested_possessions,
        contested_marks=contested_marks,
        marks_inside=marks_inside,
        one_percenters=one_percenters,
        bounces=bounces,
        goal_assists=goal_assists,
        percentage_played=percentage_played,
        birth_date=birth_date,
        species=str(Species.HUMAN),
        handicap_weight=None,
        father=None,
        sex=str(Sex.MALE),
        age=None if birth_date is None else relativedelta(birth_date, dt).years,
        starting_position=None,
        weight=weight,
        birth_address=None,
        owner=None,
        seconds_played=None,
        three_point_field_goals=None,
        three_point_field_goals_attempted=None,
        free_throws=None,
        free_throws_attempted=None,
        defensive_rebounds=None,
        steals=None,
        blocks=None,
        personal_fouls=None,
        points=None,
        game_score=None,
        point_differential=None,
        version=version,
        height=height,
        colleges=[],
        headshot=None,
        forced_fumbles=None,
        fumbles_recovered=None,
        fumbles_recovered_yards=None,
        fumbles_touchdowns=None,
        offensive_two_point_returns=None,
        offensive_fumbles_touchdowns=None,
        defensive_fumbles_touchdowns=None,
        average_gain=None,
        completion_percentage=None,
        completions=None,
        espn_quarterback_rating=None,
        interception_percentage=None,
        interceptions=None,
        long_passing=None,
        misc_yards=None,
        net_passing_yards=None,
        net_total_yards=None,
        passing_attempts=None,
        passing_big_plays=None,
        passing_first_downs=None,
        passing_fumbles=None,
        passing_fumbles_lost=None,
        passing_touchdown_percentage=None,
        passing_touchdowns=None,
        passing_yards=None,
        passing_yards_after_catch=None,
        quarterback_rating=None,
        sacks=None,
        passing_yards_at_catch=None,
        sacks_yards_lost=None,
        net_passing_attempts=None,
        total_offensive_plays=None,
        total_points=None,
        total_touchdowns=None,
        total_yards=None,
        total_yards_from_scrimmage=None,
        two_point_pass=None,
        two_point_pass_attempt=None,
        yards_per_completion=None,
        yards_per_pass_attempt=None,
        net_yards_per_pass_attempt=None,
        long_rushing=None,
        rushing_attempts=None,
        rushing_big_plays=None,
        rushing_first_downs=None,
        rushing_fumbles=None,
        rushing_fumbles_lost=None,
        rushing_touchdowns=None,
        rushing_yards=None,
        stuffs=None,
        stuff_yards_lost=None,
        two_point_rush=None,
        two_point_rush_attempts=None,
        yards_per_rush_attempt=None,
        espn_widereceiver=None,
        long_reception=None,
        receiving_big_plays=None,
        receiving_first_downs=None,
        receiving_fumbles=None,
        receiving_fumbles_lost=None,
        receiving_targets=None,
        receiving_touchdowns=None,
        receiving_yards=None,
        receiving_yards_after_catch=None,
        receiving_yards_at_catch=None,
        receptions=None,
        two_point_receptions=None,
        two_point_reception_attempts=None,
        yards_per_reception=None,
        assist_tackles=None,
        average_interception_yards=None,
        average_sack_yards=None,
        average_stuff_yards=None,
        blocked_field_goal_touchdowns=None,
        blocked_punt_touchdowns=None,
        defensive_touchdowns=None,
        hurries=None,
        kicks_blocked=None,
        long_interception=None,
        misc_touchdowns=None,
        passes_batted_down=None,
        passes_defended=None,
        quarterback_hits=None,
        sacks_assisted=None,
        sacks_unassisted=None,
        sacks_yards=None,
        safeties=None,
        solo_tackles=None,
        stuff_yards=None,
        tackles_for_loss=None,
        tackles_yards_lost=None,
        yards_allowed=None,
        points_allowed=None,
        one_point_safeties_made=None,
        missed_field_goal_return_td=None,
        blocked_punt_ez_rec_td=None,
        interception_touchdowns=None,
        interception_yards=None,
        average_kickoff_return_yards=None,
        average_kickoff_yards=None,
        extra_point_attempts=None,
        extra_point_percentage=None,
        extra_point_blocked=None,
        extra_points_blocked_percentage=None,
        extra_points_made=None,
        fair_catches=None,
        fair_catch_percentage=None,
        field_goal_attempts_max_19_yards=None,
        field_goal_attempts_max_29_yards=None,
        field_goal_attempts_max_39_yards=None,
        field_goal_attempts_max_49_yards=None,
        field_goal_attempts_max_59_yards=None,
        field_goal_attempts_max_99_yards=None,
        field_goal_attempts_above_50_yards=None,
        field_goal_attempt_yards=None,
        field_goals_blocked=None,
        field_goals_blocked_percentage=None,
        field_goals_made=None,
        field_goals_made_max_19_yards=None,
        field_goals_made_max_29_yards=None,
        field_goals_made_max_39_yards=None,
        field_goals_made_max_49_yards=None,
        field_goals_made_max_59_yards=None,
        field_goals_made_max_99_yards=None,
        field_goals_made_above_50_yards=None,
        field_goals_made_yards=None,
        field_goals_missed_yards=None,
        kickoff_out_of_bounds=None,
        kickoff_returns=None,
        kickoff_returns_touchdowns=None,
        kickoff_return_yards=None,
        kickoffs=None,
        kickoff_yards=None,
        long_field_goal_attempt=None,
        long_field_goal_made=None,
        long_kickoff=None,
        total_kicking_points=None,
        touchback_percentage=None,
        touchbacks=None,
        defensive_fumble_returns=None,
        defensive_fumble_return_yards=None,
        fumble_recoveries=None,
        fumble_recovery_yards=None,
        kick_return_fair_catches=None,
        kick_return_fair_catch_percentage=None,
        kick_return_fumbles=None,
        kick_return_fumbles_lost=None,
        kick_returns=None,
        kick_return_touchdowns=None,
        kick_return_yards=None,
        long_kick_return=None,
        long_punt_return=None,
        misc_fumble_returns=None,
        misc_fumble_return_yards=None,
        opposition_fumble_recoveries=None,
        opposition_fumble_recovery_yards=None,
        opposition_special_team_fumble_returns=None,
        opposition_special_team_fumble_return_yards=None,
        punt_return_fair_catches=None,
        punt_return_fair_catch_percentage=None,
        punt_return_fumbles=None,
        punt_return_fumbles_lost=None,
        punt_returns=None,
        punt_returns_started_inside_the_10=None,
        punt_returns_started_inside_the_20=None,
        punt_return_touchdowns=None,
        special_team_fumble_returns=None,
        yards_per_kick_return=None,
        yards_per_punt_return=None,
        yards_per_return=None,
        average_punt_return_yards=None,
        gross_average_punt_yards=None,
        long_punt=None,
        net_average_punt_yards=None,
        punts=None,
        punts_blocked=None,
        punts_blocked_percentage=None,
        punts_inside_10=None,
        punts_inside_10_percentage=None,
        punts_inside_20=None,
        punts_inside_20_percentage=None,
        punts_over_50=None,
        punt_yards=None,
        defensive_points=None,
        misc_points=None,
        return_touchdowns=None,
        total_two_point_conversions=None,
        passing_touchdowns_9_yards=None,
        passing_touchdowns_19_yards=None,
        passing_touchdowns_29_yards=None,
        passing_touchdowns_39_yards=None,
        passing_touchdowns_49_yards=None,
        passing_touchdowns_above_50_yards=None,
        receiving_touchdowns_9_yards=None,
        receiving_touchdowns_19_yards=None,
        receiving_touchdowns_29_yards=None,
        receiving_touchdowns_39_yards=None,
        punt_return_yards=None,
        receiving_touchdowns_49_yards=None,
        receiving_touchdowns_above_50_yards=None,
        rushing_touchdowns_9_yards=None,
        rushing_touchdowns_19_yards=None,
        rushing_touchdowns_29_yards=None,
        rushing_touchdowns_39_yards=None,
        rushing_touchdowns_49_yards=None,
        rushing_touchdowns_above_50_yards=None,
        penalties_in_minutes=None,
        even_strength_goals=None,
        power_play_goals=None,
        short_handed_goals=None,
        game_winning_goals=None,
        even_strength_assists=None,
        power_play_assists=None,
        short_handed_assists=None,
        shots_on_goal=None,
        shooting_percentage=None,
        shifts=None,
        time_on_ice=None,
        decision=None,
        goals_against=None,
        shots_against=None,
        saves=None,
        save_percentage=None,
        shutouts=None,
        individual_corsi_for_events=None,
        on_shot_ice_for_events=None,
        on_shot_ice_against_events=None,
        corsi_for_percentage=None,
        relative_corsi_for_percentage=None,
        offensive_zone_starts=None,
        defensive_zone_starts=None,
        offensive_zone_start_percentage=None,
        hits=None,
        true_shooting_percentage=None,
        at_bats=None,
        runs_scored=None,
        runs_batted_in=None,
        bases_on_balls=None,
        strikeouts=None,
        plate_appearances=None,
        hits_at_bats=None,
        obp=None,
        slg=None,
        ops=None,
        pitches=None,
        strikes=None,
        win_probability_added=None,
        average_leverage_index=None,
        wpa_plus=None,
        wpa_minus=None,
        cwpa=None,
        acli=None,
        re24=None,
        putouts=None,
        innings_pitched=None,
        earned_runs=None,
        home_runs=None,
        era=None,
        batters_faced=None,
        strikes_by_contact=None,
        strikes_swinging=None,
        strikes_looking=None,
        ground_balls=None,
        fly_balls=None,
        line_drives=None,
        inherited_runners=None,
        inherited_scores=None,
        effective_field_goal_percentage=None,
        penalty_kicks_made=None,
        penalty_kicks_attempted=None,
        shots_total=None,
        shots_on_target=None,
        yellow_cards=None,
        red_cards=None,
        touches=None,
        expected_goals=None,
        non_penalty_expected_goals=None,
        expected_assisted_goals=None,
        shot_creating_actions=None,
        goal_creating_actions=None,
        passes_completed=None,
        passes_attempted=None,
        pass_completion=None,
        progressive_passes=None,
        carries=None,
        progressive_carries=None,
        take_ons_attempted=None,
        successful_take_ons=None,
        total_passing_distance=None,
        progressive_passing_distance=None,
        passes_completed_short=None,
        passes_attempted_short=None,
        pass_completion_short=None,
        passes_completed_medium=None,
        passes_attempted_medium=None,
        pass_completion_medium=None,
        passes_completed_long=None,
        passes_attempted_long=None,
        pass_completion_long=None,
        expected_assists=None,
        key_passes=None,
        passes_into_final_third=None,
        passes_into_penalty_area=None,
        crosses_into_penalty_area=None,
        live_ball_passes=None,
        dead_ball_passes=None,
        passes_from_free_kicks=None,
        through_balls=None,
        switches=None,
        crosses=None,
        throw_ins_taken=None,
        corner_kicks=None,
        inswinging_corner_kicks=None,
        outswinging_corner_kicks=None,
        straight_corner_kicks=None,
        passes_offside=None,
        passes_blocked=None,
        tackles_won=None,
        tackles_in_defensive_third=None,
        tackles_in_middle_third=None,
        tackles_in_attacking_third=None,
        dribblers_tackled=None,
        dribbles_challenged=None,
        percent_of_dribblers_tackled=None,
        challenges_lost=None,
        shots_blocked=None,
        tackles_plus_interceptions=None,
        errors=None,
        touches_in_defensive_penalty_area=None,
        touches_in_defensive_third=None,
        touches_in_middle_third=None,
        touches_in_attacking_third=None,
        touches_in_attacking_penalty_area=None,
        live_ball_touches=None,
        successful_take_on_percentage=None,
        times_tackled_during_take_ons=None,
        tackled_during_take_on_percentage=None,
        total_carrying_distance=None,
        progressive_carrying_distance=None,
        carries_into_final_third=None,
        carries_into_penalty_area=None,
        miscontrols=None,
        dispossessed=None,
        passes_received=None,
        progressive_passes_received=None,
        second_yellow_card=None,
        fouls_committed=None,
        fouls_drawn=None,
        offsides=None,
        penalty_kicks_won=None,
        penalty_kicks_conceded=None,
        own_goals=None,
        ball_recoveries=None,
        aerials_won=None,
        aerials_lost=None,
        percentage_of_aerials_won=None,
        shots_on_target_against=None,
        post_shot_expected_goals=None,
        passes_attempted_minus_goal_kicks=None,
        throws_attempted=None,
        percentage_of_passes_that_were_launched=None,
        average_pass_length=None,
        goal_kicks_attempted=None,
        percentage_of_goal_kicks_that_were_launched=None,
        average_goal_kick_length=None,
        crosses_faced=None,
        crosses_stopped=None,
        percentage_crosses_stopped=None,
        defensive_actions_outside_penalty_area=None,
        average_distance_of_defensive_actions=None,
        three_point_attempt_rate=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_afl_afltables_player_model(
    player_url: str,
    jersey: str,
    kicks: int | None,
    name: str,
    marks: int | None,
    handballs: int | None,
    disposals: int | None,
    goals: int | None,
    behinds: int | None,
    hit_outs: int | None,
    tackles: int | None,
    rebounds: int | None,
    insides: int | None,
    clearances: int | None,
    clangers: int | None,
    free_kicks_for: int | None,
    free_kicks_against: int | None,
    brownlow_votes: int | None,
    contested_possessions: int | None,
    uncontested_possessions: int | None,
    contested_marks: int | None,
    marks_inside: int | None,
    one_percenters: int | None,
    bounces: int | None,
    goal_assists: int | None,
    percentage_played: float | None,
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    version: str,
) -> PlayerModel:
    return _create_afl_afltables_player_model(
        player_url=player_url,
        jersey=jersey,
        kicks=kicks,
        name=name,
        marks=marks,
        handballs=handballs,
        disposals=disposals,
        goals=goals,
        behinds=behinds,
        hit_outs=hit_outs,
        tackles=tackles,
        rebounds=rebounds,
        insides=insides,
        clearances=clearances,
        clangers=clangers,
        free_kicks_for=free_kicks_for,
        free_kicks_against=free_kicks_against,
        brownlow_votes=brownlow_votes,
        contested_possessions=contested_possessions,
        uncontested_possessions=uncontested_possessions,
        contested_marks=contested_marks,
        marks_inside=marks_inside,
        one_percenters=one_percenters,
        bounces=bounces,
        goal_assists=goal_assists,
        percentage_played=percentage_played,
        session=session,
        dt=dt,
        version=version,
    )


def create_afl_afltables_player_model(
    player_url: str,
    jersey: str,
    kicks: int | None,
    session: requests_cache.CachedSession,
    name: str,
    marks: int | None,
    handballs: int | None,
    disposals: int | None,
    goals: int | None,
    behinds: int | None,
    hit_outs: int | None,
    tackles: int | None,
    rebounds: int | None,
    insides: int | None,
    clearances: int | None,
    clangers: int | None,
    free_kicks_for: int | None,
    free_kicks_against: int | None,
    brownlow_votes: int | None,
    contested_possessions: int | None,
    uncontested_possessions: int | None,
    contested_marks: int | None,
    marks_inside: int | None,
    one_percenters: int | None,
    bounces: int | None,
    goal_assists: int | None,
    percentage_played: float | None,
    dt: datetime.datetime,
) -> PlayerModel:
    """Create a player model from AFL Tables."""
    if not pytest_is_running.is_running():
        return _cached_create_afl_afltables_player_model(
            player_url=player_url,
            jersey=jersey,
            kicks=kicks,
            name=name,
            marks=marks,
            handballs=handballs,  # pyright: ignore
            disposals=disposals,
            goals=goals,
            behinds=behinds,
            hit_outs=hit_outs,
            tackles=tackles,
            rebounds=rebounds,
            insides=insides,
            clearances=clearances,
            clangers=clangers,
            free_kicks_for=free_kicks_for,
            free_kicks_against=free_kicks_against,
            brownlow_votes=brownlow_votes,
            contested_possessions=contested_possessions,
            uncontested_possessions=uncontested_possessions,
            contested_marks=contested_marks,
            marks_inside=marks_inside,
            one_percenters=one_percenters,
            bounces=bounces,
            goal_assists=goal_assists,
            percentage_played=percentage_played,
            session=session,
            dt=dt,
            version=VERSION,
        )
    with session.cache_disabled():
        return _create_afl_afltables_player_model(
            player_url=player_url,
            jersey=jersey,
            kicks=kicks,
            name=name,
            marks=marks,
            handballs=handballs,
            disposals=disposals,
            goals=goals,
            behinds=behinds,
            hit_outs=hit_outs,
            tackles=tackles,
            rebounds=rebounds,
            insides=insides,
            clearances=clearances,
            clangers=clangers,
            free_kicks_for=free_kicks_for,
            free_kicks_against=free_kicks_against,
            brownlow_votes=brownlow_votes,
            contested_possessions=contested_possessions,
            uncontested_possessions=uncontested_possessions,
            contested_marks=contested_marks,
            marks_inside=marks_inside,
            one_percenters=one_percenters,
            bounces=bounces,
            goal_assists=goal_assists,
            percentage_played=percentage_played,
            session=session,
            dt=dt,
            version=VERSION,
        )



"""ESPN team model."""

# pylint: disable=too-many-arguments,too-many-locals,duplicate-code

import datetime
import json
import logging
from typing import Any

import pytest_is_running
import requests_cache

from ...cache import MEMORY
from ..google.google_news_model import create_google_news_models
from ..league import League
from ..odds_model import OddsModel
from ..team_model import VERSION, TeamModel
from ..x.x_social_model import create_x_social_model
from .espn_coach_model import create_espn_coach_model
from .espn_player_model import create_espn_player_model

ID_KEY = "id"


def _create_espn_team_model(
    session: requests_cache.CachedSession,
    team: dict[str, Any],
    roster_dict: dict[str, Any],
    odds: list[OddsModel],
    score_dict: dict[str, Any],
    dt: datetime.datetime,
    league: League,
    positions_validator: dict[str, str],
    version: str,
) -> TeamModel:
    identifier = team[ID_KEY]
    name = team.get("name", team.get("fullName", team.get("displayName")))
    if name is None:
        raise ValueError("name is null")
    location = team["location"]
    players = []
    for entity in roster_dict.get("entries", []):
        player = create_espn_player_model(session, entity, dt, positions_validator)
        players.append(player)
    points = score_dict["value"]
    coaches_urls = []
    if "coaches" in team:
        try:
            coaches_response = session.get(team["coaches"]["$ref"])
        except KeyError:
            logging.warning("coaches missing for team %s", json.dumps(team))
            raise
        coaches_response.raise_for_status()
        coaches_urls = [x["$ref"] for x in coaches_response.json()["items"]]
    return TeamModel(
        identifier=identifier,
        name=name,
        location=location,
        players=players,
        odds=odds,
        points=points,
        ladder_rank=None,
        news=create_google_news_models(name, session, dt, league),
        social=create_x_social_model(identifier, session, dt),
        field_goals=None,
        coaches=[create_espn_coach_model(session, dt, x) for x in coaches_urls],
        lbw=None,
        end_dt=None,
        version=version,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_espn_team_model(
    session: requests_cache.CachedSession,
    team: dict[str, Any],
    roster_dict: dict[str, Any],
    odds: list[OddsModel],
    score_dict: dict[str, Any],
    dt: datetime.datetime,
    league: League,
    positions_validator: dict[str, str],
    version: str,
) -> TeamModel:
    return _create_espn_team_model(
        session=session,
        team=team,
        roster_dict=roster_dict,
        odds=odds,
        score_dict=score_dict,
        dt=dt,
        league=league,
        positions_validator=positions_validator,
        version=version,
    )


def create_espn_team_model(
    session: requests_cache.CachedSession,
    team: dict[str, Any],
    roster_dict: dict[str, Any],
    odds: list[OddsModel],
    score_dict: dict[str, Any],
    dt: datetime.datetime,
    league: League,
    positions_validator: dict[str, str],
) -> TeamModel:
    """Create team model from ESPN."""
    if (
        not pytest_is_running.is_running()
        and dt.date() < datetime.datetime.today().date() - datetime.timedelta(days=7)
    ):
        return _cached_create_espn_team_model(
            session=session,
            team=team,
            roster_dict=roster_dict,
            odds=odds,
            score_dict=score_dict,
            dt=dt,
            league=league,
            positions_validator=positions_validator,
            version=VERSION,
        )
    with session.cache_disabled():
        return _create_espn_team_model(
            session=session,
            team=team,
            roster_dict=roster_dict,
            odds=odds,
            score_dict=score_dict,
            dt=dt,
            league=league,
            positions_validator=positions_validator,
            version=VERSION,
        )



"""Tests for the player model class."""
import unittest

from sportsball.data.player_model import PlayerModel, VERSION
from sportsball.data.species import Species


class TestPlayerModel(unittest.TestCase):

    def setUp(self):
        self._player_model = PlayerModel(
            identifier="a",
            jersey="45",
            kicks=None,
            fumbles=None,
            fumbles_lost=None,
            field_goals=None,
            field_goals_attempted=10,
            offensive_rebounds=None,
            assists=None,
            turnovers=None,
            name="James Jones",
            marks=None,
            handballs=None,
            disposals=None,
            goals=None,
            behinds=None,
            hit_outs=None,
            tackles=None,
            rebounds=None,
            insides=None,
            clearances=None,
            clangers=None,
            free_kicks_for=None,
            free_kicks_against=None,
            brownlow_votes=None,
            contested_possessions=None,
            uncontested_possessions=None,
            contested_marks=None,
            marks_inside=None,
            one_percenters=None,
            bounces=None,
            goal_assists=None,
            percentage_played=None,
            birth_date=None,
            age=None,
            species=str(Species.HUMAN),
            handicap_weight=None,
            father=None,
            starting_position=None,
            weight=None,
            birth_address=None,
            owner=None,
            seconds_played=None,
            three_point_field_goals=None,
            three_point_field_goals_attempted=None,
            free_throws=None,
            free_throws_attempted=None,
            defensive_rebounds=None,
            steals=None,
            blocks=None,
            personal_fouls=None,
            points=None,
            game_score=None,
            point_differential=None,
            version=VERSION,
            height=None,
            colleges=[],
            headshot=None,
            forced_fumbles=None,
            fumbles_recovered=None,
            fumbles_recovered_yards=None,
            fumbles_touchdowns=None,
            offensive_two_point_returns=None,
            offensive_fumbles_touchdowns=None,
            defensive_fumbles_touchdowns=None,
            average_gain=None,
            completion_percentage=None,
            completions=None,
            espn_quarterback_rating=None,
            interception_percentage=None,
            interceptions=None,
            long_passing=None,
            misc_yards=None,
            net_passing_yards=None,
            net_total_yards=None,
            passing_attempts=None,
            passing_big_plays=None,
            passing_first_downs=None,
            passing_fumbles=None,
            passing_fumbles_lost=None,
            passing_touchdown_percentage=None,
            passing_touchdowns=None,
            passing_yards=None,
            passing_yards_after_catch=None,
            quarterback_rating=None,
            sacks=None,
            passing_yards_at_catch=None,
            sacks_yards_lost=None,
            net_passing_attempts=None,
            total_offensive_plays=None,
            total_points=None,
            total_touchdowns=None,
            total_yards=None,
            total_yards_from_scrimmage=None,
            two_point_pass=None,
            two_point_pass_attempt=None,
            yards_per_completion=None,
            yards_per_pass_attempt=None,
            net_yards_per_pass_attempt=None,
            long_rushing=None,
            rushing_attempts=None,
            rushing_big_plays=None,
            rushing_first_downs=None,
            rushing_fumbles=None,
            rushing_fumbles_lost=None,
            rushing_touchdowns=None,
            rushing_yards=None,
            stuffs=None,
            stuff_yards_lost=None,
            two_point_rush=None,
            two_point_rush_attempts=None,
            yards_per_rush_attempt=None,
            espn_widereceiver=None,
            long_reception=None,
            receiving_big_plays=None,
            receiving_first_downs=None,
            receiving_fumbles=None,
            receiving_fumbles_lost=None,
            receiving_targets=None,
            receiving_touchdowns=None,
            receiving_yards=None,
            receiving_yards_after_catch=None,
            receiving_yards_at_catch=None,
            receptions=None,
            two_point_receptions=None,
            two_point_reception_attempts=None,
            yards_per_reception=None,
            assist_tackles=None,
            average_interception_yards=None,
            average_sack_yards=None,
            average_stuff_yards=None,
            blocked_field_goal_touchdowns=None,
            blocked_punt_touchdowns=None,
            defensive_touchdowns=None,
            hurries=None,
            kicks_blocked=None,
            long_interception=None,
            misc_touchdowns=None,
            passes_batted_down=None,
            passes_defended=None,
            quarterback_hits=None,
            sacks_assisted=None,
            sacks_unassisted=None,
            sacks_yards=None,
            safeties=None,
            solo_tackles=None,
            stuff_yards=None,
            tackles_for_loss=None,
            tackles_yards_lost=None,
            yards_allowed=None,
            points_allowed=None,
            one_point_safeties_made=None,
            missed_field_goal_return_td=None,
            blocked_punt_ez_rec_td=None,
            interception_touchdowns=None,
            interception_yards=None,
            average_kickoff_return_yards=None,
            average_kickoff_yards=None,
            extra_point_attempts=None,
            extra_point_percentage=None,
            extra_point_blocked=None,
            extra_points_blocked_percentage=None,
            extra_points_made=None,
            fair_catches=None,
            fair_catch_percentage=None,
            field_goal_attempts_max_19_yards=None,
            field_goal_attempts_max_29_yards=None,
            field_goal_attempts_max_39_yards=None,
            field_goal_attempts_max_49_yards=None,
            field_goal_attempts_max_59_yards=None,
            field_goal_attempts_max_99_yards=None,
            field_goal_attempts_above_50_yards=None,
            field_goal_attempt_yards=None,
            field_goals_blocked=None,
            field_goals_blocked_percentage=None,
            field_goals_made=None,
            field_goals_made_max_19_yards=None,
            field_goals_made_max_29_yards=None,
            field_goals_made_max_39_yards=None,
            field_goals_made_max_49_yards=None,
            field_goals_made_max_59_yards=None,
            field_goals_made_max_99_yards=None,
            field_goals_made_above_50_yards=None,
            field_goals_made_yards=None,
            field_goals_missed_yards=None,
            kickoff_out_of_bounds=None,
            kickoff_returns=None,
            kickoff_returns_touchdowns=None,
            kickoff_return_yards=None,
            kickoffs=None,
            kickoff_yards=None,
            long_field_goal_attempt=None,
            long_field_goal_made=None,
            long_kickoff=None,
            total_kicking_points=None,
            touchback_percentage=None,
            touchbacks=None,
            defensive_fumble_returns=None,
            defensive_fumble_return_yards=None,
            fumble_recoveries=None,
            fumble_recovery_yards=None,
            kick_return_fair_catches=None,
            kick_return_fair_catch_percentage=None,
            kick_return_fumbles=None,
            kick_return_fumbles_lost=None,
            kick_returns=None,
            kick_return_touchdowns=None,
            kick_return_yards=None,
            long_kick_return=None,
            long_punt_return=None,
            misc_fumble_returns=None,
            misc_fumble_return_yards=None,
            opposition_fumble_recoveries=None,
            opposition_fumble_recovery_yards=None,
            opposition_special_team_fumble_returns=None,
            opposition_special_team_fumble_return_yards=None,
            punt_return_fair_catches=None,
            punt_return_fair_catch_percentage=None,
            punt_return_fumbles=None,
            punt_return_fumbles_lost=None,
            punt_returns=None,
            punt_returns_started_inside_the_10=None,
            punt_returns_started_inside_the_20=None,
            punt_return_touchdowns=None,
            special_team_fumble_returns=None,
            yards_per_kick_return=None,
            yards_per_punt_return=None,
            yards_per_return=None,
            average_punt_return_yards=None,
            gross_average_punt_yards=None,
            long_punt=None,
            net_average_punt_yards=None,
            punts=None,
            punts_blocked=None,
            punts_blocked_percentage=None,
            punts_inside_10=None,
            punts_inside_10_percentage=None,
            punts_inside_20=None,
            punts_inside_20_percentage=None,
            punts_over_50=None,
            punt_yards=None,
            defensive_points=None,
            misc_points=None,
            return_touchdowns=None,
            total_two_point_conversions=None,
            passing_touchdowns_9_yards=None,
            passing_touchdowns_19_yards=None,
            passing_touchdowns_29_yards=None,
            passing_touchdowns_39_yards=None,
            passing_touchdowns_49_yards=None,
            passing_touchdowns_above_50_yards=None,
            receiving_touchdowns_9_yards=None,
            receiving_touchdowns_19_yards=None,
            receiving_touchdowns_29_yards=None,
            receiving_touchdowns_39_yards=None,
            punt_return_yards=None,
            receiving_touchdowns_49_yards=None,
            receiving_touchdowns_above_50_yards=None,
            rushing_touchdowns_9_yards=None,
            rushing_touchdowns_19_yards=None,
            rushing_touchdowns_29_yards=None,
            rushing_touchdowns_39_yards=None,
            rushing_touchdowns_49_yards=None,
            rushing_touchdowns_above_50_yards=None,
            penalties_in_minutes=None,
            even_strength_goals=None,
            power_play_goals=None,
            short_handed_goals=None,
            game_winning_goals=None,
            even_strength_assists=None,
            power_play_assists=None,
            short_handed_assists=None,
            shots_on_goal=None,
            shooting_percentage=None,
            shifts=None,
            time_on_ice=None,
            decision=None,
            goals_against=None,
            shots_against=None,
            saves=None,
            save_percentage=None,
            shutouts=None,
            individual_corsi_for_events=None,
            on_shot_ice_for_events=None,
            on_shot_ice_against_events=None,
            corsi_for_percentage=None,
            relative_corsi_for_percentage=None,
            offensive_zone_starts=None,
            defensive_zone_starts=None,
            offensive_zone_start_percentage=None,
            hits=None,
            true_shooting_percentage=None,
            at_bats=None,
            runs_scored=None,
            runs_batted_in=None,
            bases_on_balls=None,
            strikeouts=None,
            plate_appearances=None,
            hits_at_bats=None,
            obp=None,
            slg=None,
            ops=None,
            pitches=None,
            strikes=None,
            win_probability_added=None,
            average_leverage_index=None,
            wpa_plus=None,
            wpa_minus=None,
            cwpa=None,
            acli=None,
            re24=None,
            putouts=None,
            innings_pitched=None,
            earned_runs=None,
            home_runs=None,
            era=None,
            batters_faced=None,
            strikes_by_contact=None,
            strikes_swinging=None,
            strikes_looking=None,
            ground_balls=None,
            fly_balls=None,
            line_drives=None,
            inherited_runners=None,
            inherited_scores=None,
            effective_field_goal_percentage=None,
            penalty_kicks_made=None,
            penalty_kicks_attempted=None,
            shots_total=None,
            shots_on_target=None,
            yellow_cards=None,
            red_cards=None,
            touches=None,
            expected_goals=None,
            non_penalty_expected_goals=None,
            expected_assisted_goals=None,
            shot_creating_actions=None,
            goal_creating_actions=None,
            passes_completed=None,
            passes_attempted=None,
            pass_completion=None,
            progressive_passes=None,
            carries=None,
            progressive_carries=None,
            take_ons_attempted=None,
            successful_take_ons=None,
            total_passing_distance=None,
            progressive_passing_distance=None,
            passes_completed_short=None,
            passes_attempted_short=None,
            pass_completion_short=None,
            passes_completed_medium=None,
            passes_attempted_medium=None,
            pass_completion_medium=None,
            passes_completed_long=None,
            passes_attempted_long=None,
            pass_completion_long=None,
            expected_assists=None,
            key_passes=None,
            passes_into_final_third=None,
            passes_into_penalty_area=None,
            crosses_into_penalty_area=None,
            live_ball_passes=None,
            dead_ball_passes=None,
            passes_from_free_kicks=None,
            through_balls=None,
            switches=None,
            crosses=None,
            throw_ins_taken=None,
            corner_kicks=None,
            inswinging_corner_kicks=None,
            outswinging_corner_kicks=None,
            straight_corner_kicks=None,
            passes_offside=None,
            passes_blocked=None,
            tackles_won=None,
            tackles_in_defensive_third=None,
            tackles_in_middle_third=None,
            tackles_in_attacking_third=None,
            dribblers_tackled=None,
            dribbles_challenged=None,
            percent_of_dribblers_tackled=None,
            challenges_lost=None,
            shots_blocked=None,
            tackles_plus_interceptions=None,
            errors=None,
            touches_in_defensive_penalty_area=None,
            touches_in_defensive_third=None,
            touches_in_middle_third=None,
            touches_in_attacking_third=None,
            touches_in_attacking_penalty_area=None,
            live_ball_touches=None,
            successful_take_on_percentage=None,
            times_tackled_during_take_ons=None,
            tackled_during_take_on_percentage=None,
            total_carrying_distance=None,
            progressive_carrying_distance=None,
            carries_into_final_third=None,
            carries_into_penalty_area=None,
            miscontrols=None,
            dispossessed=None,
            passes_received=None,
            progressive_passes_received=None,
            second_yellow_card=None,
            fouls_committed=None,
            fouls_drawn=None,
            offsides=None,
            penalty_kicks_won=None,
            penalty_kicks_conceded=None,
            own_goals=None,
            ball_recoveries=None,
            aerials_won=None,
            aerials_lost=None,
            percentage_of_aerials_won=None,
            shots_on_target_against=None,
            post_shot_expected_goals=None,
            passes_attempted_minus_goal_kicks=None,
            throws_attempted=None,
            percentage_of_passes_that_were_launched=None,
            average_pass_length=None,
            goal_kicks_attempted=None,
            percentage_of_goal_kicks_that_were_launched=None,
            average_goal_kick_length=None,
            crosses_faced=None,
            crosses_stopped=None,
            percentage_crosses_stopped=None,
            defensive_actions_outside_penalty_area=None,
            average_distance_of_defensive_actions=None,
            three_point_attempt_rate=None,
        )

    def test_field_goals_attempted(self):
        self.assertEqual(self._player_model.field_goals_attempted, 10)



"""Find the args used by the main CLI."""

import argparse

from .data.league import League
from .loglevel import LogLevel

STDOUT_FILE = "-"


def parse_args() -> argparse.Namespace:
    """Create the args based on the CLI inputs."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--league",
        choices=list(League),
        help="The league to fetch data for.",
    )
    parser.add_argument(
        "--loglevel",
        default=LogLevel.INFO,
        choices=list(LogLevel),
        help="The loglevel to display logs at..",
        required=False,
    )
    parser.add_argument(
        "--leaguemodel",
        required=False,
        help="The league model to use.",
    )
    parser.add_argument(
        "file",
        default=STDOUT_FILE,
        help="The file to write the output to (- if to stdout).",
    )
    return parser.parse_args()



"""NFL combined league model."""

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..aussportsbetting.nfl_aussportsbetting_league_model import \
    NFLAusSportsBettingLeagueModel
from ..espn.nfl_espn_league_model import NFLESPNLeagueModel
from ..oddsportal.nfl_oddsportal_league_model import NFLOddsPortalLeagueModel
from ..sportsdb.nfl_sportsdb_league_model import NFLSportsDBLeagueModel

# from ..sportsreference.nfl_sportsreference_league_model import \
#    NFLSportsReferenceLeagueModel

LOS_ANGELES_RAMS = "135907"
CHICAGO_BEARS = "134938"
HOUSTAN_TEXANS = "134926"
NEW_YORK_GIANTS = "134935"
DETROIT_LIONS = "134939"
NEW_ENGLAND_PATRIOTS = "134920"
CAROLINA_PANTHERS = "134943"
PITTSBURGH_STEALERS = "134925"
MIAMI_DOLPHINS = "134919"
ATLANTA_FALCONS = "134942"
BALTIMORE_RAVENS = "134922"
PHILADELPHIA_EAGLES = "134936"
NEW_YORK_JETS = "134921"
WASHINGTON_COMMANDERS = "134937"
BUFFALO_BILLS = "134918"
MINNESOTA_VIKINGS = "134941"
LAS_VEGAS_RAIDERS = "134932"
CLEVELAND_BROWNS = "134924"
GREEN_BAY_PACKERS = "134940"
JACKSONVILLE_JAGUARS = "134928"
KANSAS_CITY_CHIEFS = "134931"
TENNESSEE_TITANS = "134929"
SAN_FRANCISCO_49ERS = "134948"
CINCINATTI_BENGALS = "134923"
TAMPA_BAY_BUCCANEERS = "134945"
LOS_ANGELES_CHARGERS = "135908"
SEATTLE_SEAHAWKS = "134949"
ARIZONA_CARDINALS = "134946"
NEW_ORLEANS_SAINTS = "134944"
INDIANAPOLIS_COLTS = "134927"
DENVER_BRONCOS = "134930"
DALLAS_COWBOYS = "134934"
AMERICAN_FOOTBALL_CONFERENCE = "31"
NATIONAL_FOOTBALL_CONFERENCE = "32"
NFL_TEAM_IDENTITY_MAP = {
    "135907": LOS_ANGELES_RAMS,
    "134938": CHICAGO_BEARS,
    "134926": HOUSTAN_TEXANS,
    "134935": NEW_YORK_GIANTS,
    "134939": DETROIT_LIONS,
    "134920": NEW_ENGLAND_PATRIOTS,
    "134943": CAROLINA_PANTHERS,
    "134925": PITTSBURGH_STEALERS,
    "134919": MIAMI_DOLPHINS,
    "134942": ATLANTA_FALCONS,
    "134922": BALTIMORE_RAVENS,
    "134936": PHILADELPHIA_EAGLES,
    "134921": NEW_YORK_JETS,
    "134937": WASHINGTON_COMMANDERS,
    "134918": BUFFALO_BILLS,
    "134941": MINNESOTA_VIKINGS,
    "134932": LAS_VEGAS_RAIDERS,
    "134924": CLEVELAND_BROWNS,
    "134940": GREEN_BAY_PACKERS,
    "134928": JACKSONVILLE_JAGUARS,
    "134931": KANSAS_CITY_CHIEFS,
    "134929": TENNESSEE_TITANS,
    "134948": SAN_FRANCISCO_49ERS,
    "134923": CINCINATTI_BENGALS,
    "134945": TAMPA_BAY_BUCCANEERS,
    "135908": LOS_ANGELES_CHARGERS,
    "134949": SEATTLE_SEAHAWKS,
    "134946": ARIZONA_CARDINALS,
    "134944": NEW_ORLEANS_SAINTS,
    "134927": INDIANAPOLIS_COLTS,
    "134930": DENVER_BRONCOS,
    "134934": DALLAS_COWBOYS,
    "5": CLEVELAND_BROWNS,
    "20": NEW_YORK_JETS,
    "17": NEW_ENGLAND_PATRIOTS,
    "34": HOUSTAN_TEXANS,
    "26": SEATTLE_SEAHAWKS,
    "16": MINNESOTA_VIKINGS,
    "4": CINCINATTI_BENGALS,
    "9": GREEN_BAY_PACKERS,
    "8": DETROIT_LIONS,
    "19": NEW_YORK_GIANTS,
    "15": MIAMI_DOLPHINS,
    "1": ATLANTA_FALCONS,
    "27": TAMPA_BAY_BUCCANEERS,
    "23": PITTSBURGH_STEALERS,
    "28": WASHINGTON_COMMANDERS,
    "22": ARIZONA_CARDINALS,
    "7": DENVER_BRONCOS,
    "2": BUFFALO_BILLS,
    "11": INDIANAPOLIS_COLTS,
    "3": CHICAGO_BEARS,
    "10": TENNESSEE_TITANS,
    "29": CAROLINA_PANTHERS,
    "6": DALLAS_COWBOYS,
    "30": JACKSONVILLE_JAGUARS,
    "33": BALTIMORE_RAVENS,
    "21": PHILADELPHIA_EAGLES,
    "14": LOS_ANGELES_RAMS,
    "24": LOS_ANGELES_CHARGERS,
    "18": NEW_ORLEANS_SAINTS,
    "12": KANSAS_CITY_CHIEFS,
    "13": LAS_VEGAS_RAIDERS,
    "25": SAN_FRANCISCO_49ERS,
    "31": AMERICAN_FOOTBALL_CONFERENCE,
    "32": NATIONAL_FOOTBALL_CONFERENCE,
    "134933": LOS_ANGELES_CHARGERS,
    "134947": LOS_ANGELES_RAMS,
    "135834": LOS_ANGELES_RAMS,
    "Miami Dolphins": MIAMI_DOLPHINS,
    "Pittsburgh Steelers": PITTSBURGH_STEALERS,
    "San Francisco 49ers": SAN_FRANCISCO_49ERS,
    "Arizona Cardinals": ARIZONA_CARDINALS,
    "Atlanta Falcons": ATLANTA_FALCONS,
    "Carolina Panthers": CAROLINA_PANTHERS,
    "New Orleans Saints": NEW_ORLEANS_SAINTS,
    "Cleveland Browns": CLEVELAND_BROWNS,
    "Seattle Seahawks": SEATTLE_SEAHAWKS,
    "Detroit Lions": DETROIT_LIONS,
    "Chicago Bears": CHICAGO_BEARS,
    "Green Bay Packers": GREEN_BAY_PACKERS,
    "Philadelphia Eagles": PHILADELPHIA_EAGLES,
    "Houston Texans": HOUSTAN_TEXANS,
    "Dallas Cowboys": DALLAS_COWBOYS,
    "Jacksonville Jaguars": JACKSONVILLE_JAGUARS,
    "Cincinnati Bengals": CINCINATTI_BENGALS,
    "Kansas City Chiefs": KANSAS_CITY_CHIEFS,
    "Buffalo Bills": BUFFALO_BILLS,
    "New England Patriots": NEW_ENGLAND_PATRIOTS,
    "Indianapolis Colts": INDIANAPOLIS_COLTS,
    "New York Giants": NEW_YORK_GIANTS,
    "Denver Broncos": DENVER_BRONCOS,
    "St. Louis Rams": LOS_ANGELES_RAMS,
    "Baltimore Ravens": BALTIMORE_RAVENS,
    "Tampa Bay Buccaneers": TAMPA_BAY_BUCCANEERS,
    "New York Jets": NEW_YORK_JETS,
    "Tennessee Titans": TENNESSEE_TITANS,
    "San Diego Chargers": LOS_ANGELES_CHARGERS,
    "Oakland Raiders": LAS_VEGAS_RAIDERS,
    "Minnesota Vikings": MINNESOTA_VIKINGS,
    "Washington Redskins": WASHINGTON_COMMANDERS,
    "Los Angeles Rams": LOS_ANGELES_RAMS,
    "Los Angeles Chargers": LOS_ANGELES_CHARGERS,
    "Washington Football Team": WASHINGTON_COMMANDERS,
    "Las Vegas Raiders": LAS_VEGAS_RAIDERS,
    "Washington Commanders": WASHINGTON_COMMANDERS,
}
EMPOWER_FIELD = "18784"
TOM_BENSOM_HALL_OF_FAME_STADIUM = "28739"
METLIFE_STADIUM = "15812"
GILETTE_STADIUM = "15816"
ACRISURE_STADIUM = "29572"
HARD_ROCK_STADIUM = "20373"
MTT_BANK_STADIUM = "21598"
HIGHMARK_STADIUM = "30812"
US_BANK_STADIUM = "15803"
CLEVELAND_BROWNS_STADIUM = "29570"
EVERBANK_STADIUM = "29576"
NISSAN_STADIUM = "30912"
PAYCOR_STADIUM = "29569"
SOFI_STADIUM = "23720"
STATE_FARM_STADIUM = "16027"
LUCAS_OIL_STADIUM = "15807"
NRG_STADIUM = "15873"
SOLDIER_FIELD = "16092"
ARROWHEAD_STADIUM = "29575"
BANK_OF_AMERICA_STADIUM = "18776"
ALLEGIANT_STADIUM = "21293"
LEVIS_STADIUM = "16090"
MERCEDEZ_BENZ_STADIUM = "16029"
RAYMOND_JAMES_STADIUM = "15874"
LAMBEAU_FIELD = "17146"
LINCOLN_FINANCIAL_FIELD = "16094"
FORD_FIELD = "15808"
ATT_STADIUM = "15802"
LUMEN_FIELD = "19833"
CAESARS_SUPERDOME = "29571"
NEO_QUIMICA_ARENA = "16461"
FEDEX_FIELD = "29574"
TOTTENHAM_HOTSPUR_STADIUM = "16118"
WEMBLEY_STADIUM = "16163"
ALLIANZ_ARENA = "15537"
DEUTSCHE_BANK_PARK = "17590"
CAMPING_WORLD_STADIUM = "4013"
ESTADIO_AZTECA = "1850"
DIGNITY_HEALTH_SPORTS_PARK = "538"
OAKLAND_COLISEUM = "3867"
LOS_ANGELES_MEMORIAL_COLISEUM = "477"
TWICKENHAM_STADIUM = "5128"
GEORGIA_DOME = "3495"
SDCCU_STADIUM = "3932"
TRANS_WORLD_DOME = "3494"
CANDLESTICK_PARK = "3669"
ROGERS_CENTRE = "1822"
ALOHA_STADIUM = "3610"
TEXAS_STADIUM = "3954"
MOUNTAIN_AMERICA_STADIUM = "3947"
SILVERDOME = "3929"
LIBERTY_BOWL_MEMORIAL_STADIUM = "21515"
RFK_STADIUM = "18381"
VANDERBILT_STADIUM = "21572"
ROSE_BOWL = "16093"
MIAMI_ORANGE_BOWL = "20447"
STANFORD_STADIUM = "18382"
MILWAUKEE_COUNTY_STADIUM = "30850"
SHEA_STADIUM = "21813"
YALE_BOWL = "21608"
TIGER_STADIUM = "21532"
WAR_MEMORIAL_STADIUM = "21581"
YANKEE_STADIUM = "19474"
MUNICIPAL_STADIUM = "21642"
COTTON_BOWL = "18378"
KEZAR_STADIUM = "25773"
FRANKLIN_FIELD = "21659"
WRIGLEY_FIELD = "30763"
CITY_STADIUM = "19533"
IG_FIELD = "4461"
RICE_STADIUM = "21557"
INDEPENDENCE_STADIUM = "3766"
MEMORIAL_STADIUM = "16280"
NFL_VENUE_IDENTITY_MAP = {
    "18784": EMPOWER_FIELD,
    "28739": TOM_BENSOM_HALL_OF_FAME_STADIUM,
    "15812": METLIFE_STADIUM,
    "15816": GILETTE_STADIUM,
    "29572": ACRISURE_STADIUM,
    "20373": HARD_ROCK_STADIUM,
    "21598": MTT_BANK_STADIUM,
    "30812": HIGHMARK_STADIUM,
    "15803": US_BANK_STADIUM,
    "29570": CLEVELAND_BROWNS_STADIUM,
    "29576": EVERBANK_STADIUM,
    "30912": NISSAN_STADIUM,
    "29569": PAYCOR_STADIUM,
    "23720": SOFI_STADIUM,
    "16027": STATE_FARM_STADIUM,
    "15807": LUCAS_OIL_STADIUM,
    "15873": NRG_STADIUM,
    "16092": SOLDIER_FIELD,
    "29575": ARROWHEAD_STADIUM,
    "18776": BANK_OF_AMERICA_STADIUM,
    "21293": ALLEGIANT_STADIUM,
    "16090": LEVIS_STADIUM,
    "16029": MERCEDEZ_BENZ_STADIUM,
    "15874": RAYMOND_JAMES_STADIUM,
    "17146": LAMBEAU_FIELD,
    "16094": LINCOLN_FINANCIAL_FIELD,
    "15808": FORD_FIELD,
    "15802": ATT_STADIUM,
    "19833": LUMEN_FIELD,
    "29571": CAESARS_SUPERDOME,
    "16461": NEO_QUIMICA_ARENA,
    "29574": FEDEX_FIELD,
    "16118": TOTTENHAM_HOTSPUR_STADIUM,
    "16163": WEMBLEY_STADIUM,
    "15537": ALLIANZ_ARENA,
    "29573": LEVIS_STADIUM,
    "3718": TOM_BENSOM_HALL_OF_FAME_STADIUM,
    "3738": GILETTE_STADIUM,
    "3673": LUMEN_FIELD,
    "3874": PAYCOR_STADIUM,
    "3727": FORD_FIELD,
    "3948": HARD_ROCK_STADIUM,
    "3886": RAYMOND_JAMES_STADIUM,
    "3679": CLEVELAND_BROWNS_STADIUM,
    "3970": STATE_FARM_STADIUM,
    "3883": HIGHMARK_STADIUM,
    "3933": SOLDIER_FIELD,
    "3628": BANK_OF_AMERICA_STADIUM,
    "3687": ATT_STADIUM,
    "3814": MTT_BANK_STADIUM,
    "7065": SOFI_STADIUM,
    "3493": CAESARS_SUPERDOME,
    "6501": ALLEGIANT_STADIUM,
    "3806": LINCOLN_FINANCIAL_FIELD,
    "3839": METLIFE_STADIUM,
    "5348": MERCEDEZ_BENZ_STADIUM,
    "3891": NRG_STADIUM,
    "3752": ACRISURE_STADIUM,
    "3812": LUCAS_OIL_STADIUM,
    "3798": LAMBEAU_FIELD,
    "5239": US_BANK_STADIUM,
    "4738": LEVIS_STADIUM,
    "3719": FEDEX_FIELD,
    "3810": NISSAN_STADIUM,
    "3622": ARROWHEAD_STADIUM,
    "3712": EVERBANK_STADIUM,
    "3937": EMPOWER_FIELD,
    "15814": LEVIS_STADIUM,
    "17590": DEUTSCHE_BANK_PARK,
    "2455": WEMBLEY_STADIUM,
    "5534": TOTTENHAM_HOTSPUR_STADIUM,
    "1875": DEUTSCHE_BANK_PARK,
    "4013": CAMPING_WORLD_STADIUM,
    "1775": ALLIANZ_ARENA,
    "1850": ESTADIO_AZTECA,
    "21629": ARROWHEAD_STADIUM,
    "21607": PAYCOR_STADIUM,
    "538": DIGNITY_HEALTH_SPORTS_PARK,
    "3867": OAKLAND_COLISEUM,
    "477": LOS_ANGELES_MEMORIAL_COLISEUM,
    "5128": TWICKENHAM_STADIUM,
    "3495": GEORGIA_DOME,
    "3932": SDCCU_STADIUM,
    "3494": TRANS_WORLD_DOME,
    "3953": US_BANK_STADIUM,
    "3669": CANDLESTICK_PARK,
    "16": US_BANK_STADIUM,
    "1822": ROGERS_CENTRE,
    "3610": ALOHA_STADIUM,
    "3736": METLIFE_STADIUM,
    "3840": US_BANK_STADIUM,
    "3954": TEXAS_STADIUM,
    "50": EMPOWER_FIELD,
    "24": CANDLESTICK_PARK,
    "3888": LUCAS_OIL_STADIUM,
    "3947": MOUNTAIN_AMERICA_STADIUM,
    "3977": LINCOLN_FINANCIAL_FIELD,
    "3929": SILVERDOME,
    "28229": TRANS_WORLD_DOME,
    "30854": LUMEN_FIELD,
    "15804": GEORGIA_DOME,
    "28181": CAESARS_SUPERDOME,
    "18214": METLIFE_STADIUM,
    "23380": EMPOWER_FIELD,
    "18379": SILVERDOME,
    "18383": GILETTE_STADIUM,
    "25758": LINCOLN_FINANCIAL_FIELD,
    "20450": TEXAS_STADIUM,
    "20449": ACRISURE_STADIUM,
    "21559": SDCCU_STADIUM,
    "21501": MOUNTAIN_AMERICA_STADIUM,
    "3955": ACRISURE_STADIUM,
    "21515": LIBERTY_BOWL_MEMORIAL_STADIUM,
    "28192": NRG_STADIUM,
    "18381": RFK_STADIUM,
    "30845": PAYCOR_STADIUM,
    "30853": CANDLESTICK_PARK,
    "30846": PAYCOR_STADIUM,
    "20451": LOS_ANGELES_MEMORIAL_COLISEUM,
    "21572": VANDERBILT_STADIUM,
    "16093": ROSE_BOWL,
    "20447": MIAMI_ORANGE_BOWL,
    "18382": STANFORD_STADIUM,
    "30850": MILWAUKEE_COUNTY_STADIUM,
    "21813": SHEA_STADIUM,
    "20448": GILETTE_STADIUM,
    "21608": YALE_BOWL,
    "21532": TIGER_STADIUM,
    "21581": WAR_MEMORIAL_STADIUM,
    "19474": YANKEE_STADIUM,
    "21642": MUNICIPAL_STADIUM,
    "18378": COTTON_BOWL,
    "25773": KEZAR_STADIUM,
    "21659": FRANKLIN_FIELD,
    "30763": WRIGLEY_FIELD,
    "19533": CITY_STADIUM,
    "4461": IG_FIELD,
    "21557": RICE_STADIUM,
    "3766": INDEPENDENCE_STADIUM,
    "16280": MEMORIAL_STADIUM,
    "23848": SHEA_STADIUM,
    "17145": EMPOWER_FIELD,
    "30856": ATT_STADIUM,
}
NFL_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class NFLCombinedLeagueModel(CombinedLeagueModel):
    """NFL combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.NFL,
            [
                NFLESPNLeagueModel(session, position=0),
                NFLSportsDBLeagueModel(session, position=1),
                NFLAusSportsBettingLeagueModel(session, position=2),
                NFLOddsPortalLeagueModel(session, position=3),
                # NFLSportsReferenceLeagueModel(session, position=4),
            ],
            league_filter,
        )

    @classmethod
    def name(cls) -> str:
        return "nfl-combined-league-model"

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return NFL_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return NFL_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return NFL_PLAYER_IDENTITY_MAP



"""Tests for the combined team model class."""
import datetime
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.combined.combined_team_model import create_combined_team_model
from sportsball.data.team_model import TeamModel, VERSION
from sportsball.data.player_model import PlayerModel, VERSION as PLAYER_VERSION
from sportsball.data.species import Species
from sportsball.data.coach_model import CoachModel, VERSION as COACH_VERSION


class TestCombinedTeamModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_names_resolve(self):
        with requests_mock.Mocker() as m:
            names = {}
            coach_names = {}
            player_ffill = {}
            team_ffill = {}
            coach_ffill = {}
            player_models = [PlayerModel(
                identifier="1",
                jersey=None,
                kicks=None,
                fumbles=None,
                fumbles_lost=None,
                field_goals=None,
                field_goals_attempted=None,
                offensive_rebounds=None,
                assists=None,
                turnovers=None,
                name="Johnny Johnson",
                marks=None,
                handballs=None,
                disposals=None,
                goals=None,
                behinds=None,
                hit_outs=None,
                tackles=None,
                rebounds=None,
                insides=None,
                clearances=None,
                clangers=None,
                free_kicks_for=None,
                free_kicks_against=None,
                brownlow_votes=None,
                contested_possessions=None,
                uncontested_possessions=None,
                contested_marks=None,
                marks_inside=None,
                one_percenters=None,
                bounces=None,
                goal_assists=None,
                percentage_played=None,
                birth_date=None,
                age=None,
                species=str(Species.HUMAN),
                handicap_weight=None,
                father=None,
                starting_position=None,
                weight=None,
                birth_address=None,
                owner=None,
                seconds_played=None,
                three_point_field_goals=None,
                three_point_field_goals_attempted=None,
                free_throws=None,
                free_throws_attempted=None,
                defensive_rebounds=None,
                steals=None,
                blocks=None,
                personal_fouls=None,
                points=None,
                game_score=None,
                point_differential=None,
                version=PLAYER_VERSION,
                height=None,
                colleges=[],
                headshot=None,
                forced_fumbles=None,
                fumbles_recovered=None,
                fumbles_recovered_yards=None,
                fumbles_touchdowns=None,
                offensive_two_point_returns=None,
                offensive_fumbles_touchdowns=None,
                defensive_fumbles_touchdowns=None,
                average_gain=None,
                completion_percentage=None,
                completions=None,
                espn_quarterback_rating=None,
                interception_percentage=None,
                interceptions=None,
                long_passing=None,
                misc_yards=None,
                net_passing_yards=None,
                net_total_yards=None,
                passing_attempts=None,
                passing_big_plays=None,
                passing_first_downs=None,
                passing_fumbles=None,
                passing_fumbles_lost=None,
                passing_touchdown_percentage=None,
                passing_touchdowns=None,
                passing_yards=None,
                passing_yards_after_catch=None,
                quarterback_rating=None,
                sacks=None,
                passing_yards_at_catch=None,
                sacks_yards_lost=None,
                net_passing_attempts=None,
                total_offensive_plays=None,
                total_points=None,
                total_touchdowns=None,
                total_yards=None,
                total_yards_from_scrimmage=None,
                two_point_pass=None,
                two_point_pass_attempt=None,
                yards_per_completion=None,
                yards_per_pass_attempt=None,
                net_yards_per_pass_attempt=None,
                long_rushing=None,
                rushing_attempts=None,
                rushing_big_plays=None,
                rushing_first_downs=None,
                rushing_fumbles=None,
                rushing_fumbles_lost=None,
                rushing_touchdowns=None,
                rushing_yards=None,
                stuffs=None,
                stuff_yards_lost=None,
                two_point_rush=None,
                two_point_rush_attempts=None,
                yards_per_rush_attempt=None,
                espn_widereceiver=None,
                long_reception=None,
                receiving_big_plays=None,
                receiving_first_downs=None,
                receiving_fumbles=None,
                receiving_fumbles_lost=None,
                receiving_targets=None,
                receiving_touchdowns=None,
                receiving_yards=None,
                receiving_yards_after_catch=None,
                receiving_yards_at_catch=None,
                receptions=None,
                two_point_receptions=None,
                two_point_reception_attempts=None,
                yards_per_reception=None,
                assist_tackles=None,
                average_interception_yards=None,
                average_sack_yards=None,
                average_stuff_yards=None,
                blocked_field_goal_touchdowns=None,
                blocked_punt_touchdowns=None,
                defensive_touchdowns=None,
                hurries=None,
                kicks_blocked=None,
                long_interception=None,
                misc_touchdowns=None,
                passes_batted_down=None,
                passes_defended=None,
                quarterback_hits=None,
                sacks_assisted=None,
                sacks_unassisted=None,
                sacks_yards=None,
                safeties=None,
                solo_tackles=None,
                stuff_yards=None,
                tackles_for_loss=None,
                tackles_yards_lost=None,
                yards_allowed=None,
                points_allowed=None,
                one_point_safeties_made=None,
                missed_field_goal_return_td=None,
                blocked_punt_ez_rec_td=None,
                interception_touchdowns=None,
                interception_yards=None,
                average_kickoff_return_yards=None,
                average_kickoff_yards=None,
                extra_point_attempts=None,
                extra_point_percentage=None,
                extra_point_blocked=None,
                extra_points_blocked_percentage=None,
                extra_points_made=None,
                fair_catches=None,
                fair_catch_percentage=None,
                field_goal_attempts_max_19_yards=None,
                field_goal_attempts_max_29_yards=None,
                field_goal_attempts_max_39_yards=None,
                field_goal_attempts_max_49_yards=None,
                field_goal_attempts_max_59_yards=None,
                field_goal_attempts_max_99_yards=None,
                field_goal_attempts_above_50_yards=None,
                field_goal_attempt_yards=None,
                field_goals_blocked=None,
                field_goals_blocked_percentage=None,
                field_goals_made=None,
                field_goals_made_max_19_yards=None,
                field_goals_made_max_29_yards=None,
                field_goals_made_max_39_yards=None,
                field_goals_made_max_49_yards=None,
                field_goals_made_max_59_yards=None,
                field_goals_made_max_99_yards=None,
                field_goals_made_above_50_yards=None,
                field_goals_made_yards=None,
                field_goals_missed_yards=None,
                kickoff_out_of_bounds=None,
                kickoff_returns=None,
                kickoff_returns_touchdowns=None,
                kickoff_return_yards=None,
                kickoffs=None,
                kickoff_yards=None,
                long_field_goal_attempt=None,
                long_field_goal_made=None,
                long_kickoff=None,
                total_kicking_points=None,
                touchback_percentage=None,
                touchbacks=None,
                defensive_fumble_returns=None,
                defensive_fumble_return_yards=None,
                fumble_recoveries=None,
                fumble_recovery_yards=None,
                kick_return_fair_catches=None,
                kick_return_fair_catch_percentage=None,
                kick_return_fumbles=None,
                kick_return_fumbles_lost=None,
                kick_returns=None,
                kick_return_touchdowns=None,
                kick_return_yards=None,
                long_kick_return=None,
                long_punt_return=None,
                misc_fumble_returns=None,
                misc_fumble_return_yards=None,
                opposition_fumble_recoveries=None,
                opposition_fumble_recovery_yards=None,
                opposition_special_team_fumble_returns=None,
                opposition_special_team_fumble_return_yards=None,
                punt_return_fair_catches=None,
                punt_return_fair_catch_percentage=None,
                punt_return_fumbles=None,
                punt_return_fumbles_lost=None,
                punt_returns=None,
                punt_returns_started_inside_the_10=None,
                punt_returns_started_inside_the_20=None,
                punt_return_touchdowns=None,
                special_team_fumble_returns=None,
                yards_per_kick_return=None,
                yards_per_punt_return=None,
                yards_per_return=None,
                average_punt_return_yards=None,
                gross_average_punt_yards=None,
                long_punt=None,
                net_average_punt_yards=None,
                punts=None,
                punts_blocked=None,
                punts_blocked_percentage=None,
                punts_inside_10=None,
                punts_inside_10_percentage=None,
                punts_inside_20=None,
                punts_inside_20_percentage=None,
                punts_over_50=None,
                punt_yards=None,
                defensive_points=None,
                misc_points=None,
                return_touchdowns=None,
                total_two_point_conversions=None,
                passing_touchdowns_9_yards=None,
                passing_touchdowns_19_yards=None,
                passing_touchdowns_29_yards=None,
                passing_touchdowns_39_yards=None,
                passing_touchdowns_49_yards=None,
                passing_touchdowns_above_50_yards=None,
                receiving_touchdowns_9_yards=None,
                receiving_touchdowns_19_yards=None,
                receiving_touchdowns_29_yards=None,
                receiving_touchdowns_39_yards=None,
                punt_return_yards=None,
                receiving_touchdowns_49_yards=None,
                receiving_touchdowns_above_50_yards=None,
                rushing_touchdowns_9_yards=None,
                rushing_touchdowns_19_yards=None,
                rushing_touchdowns_29_yards=None,
                rushing_touchdowns_39_yards=None,
                rushing_touchdowns_49_yards=None,
                rushing_touchdowns_above_50_yards=None,
                penalties_in_minutes=None,
                even_strength_goals=None,
                power_play_goals=None,
                short_handed_goals=None,
                game_winning_goals=None,
                even_strength_assists=None,
                power_play_assists=None,
                short_handed_assists=None,
                shots_on_goal=None,
                shooting_percentage=None,
                shifts=None,
                time_on_ice=None,
                decision=None,
                goals_against=None,
                shots_against=None,
                saves=None,
                save_percentage=None,
                shutouts=None,
                individual_corsi_for_events=None,
                on_shot_ice_for_events=None,
                on_shot_ice_against_events=None,
                corsi_for_percentage=None,
                relative_corsi_for_percentage=None,
                offensive_zone_starts=None,
                defensive_zone_starts=None,
                offensive_zone_start_percentage=None,
                hits=None,
                true_shooting_percentage=None,
                at_bats=None,
                runs_scored=None,
                runs_batted_in=None,
                bases_on_balls=None,
                strikeouts=None,
                plate_appearances=None,
                hits_at_bats=None,
                obp=None,
                slg=None,
                ops=None,
                pitches=None,
                strikes=None,
                win_probability_added=None,
                average_leverage_index=None,
                wpa_plus=None,
                wpa_minus=None,
                cwpa=None,
                acli=None,
                re24=None,
                putouts=None,
                innings_pitched=None,
                earned_runs=None,
                home_runs=None,
                era=None,
                batters_faced=None,
                strikes_by_contact=None,
                strikes_swinging=None,
                strikes_looking=None,
                ground_balls=None,
                fly_balls=None,
                line_drives=None,
                inherited_runners=None,
                inherited_scores=None,
                effective_field_goal_percentage=None,
                penalty_kicks_made=None,
                penalty_kicks_attempted=None,
                shots_total=None,
                shots_on_target=None,
                yellow_cards=None,
                red_cards=None,
                touches=None,
                expected_goals=None,
                non_penalty_expected_goals=None,
                expected_assisted_goals=None,
                shot_creating_actions=None,
                goal_creating_actions=None,
                passes_completed=None,
                passes_attempted=None,
                pass_completion=None,
                progressive_passes=None,
                carries=None,
                progressive_carries=None,
                take_ons_attempted=None,
                successful_take_ons=None,
                total_passing_distance=None,
                progressive_passing_distance=None,
                passes_completed_short=None,
                passes_attempted_short=None,
                pass_completion_short=None,
                passes_completed_medium=None,
                passes_attempted_medium=None,
                pass_completion_medium=None,
                passes_completed_long=None,
                passes_attempted_long=None,
                pass_completion_long=None,
                expected_assists=None,
                key_passes=None,
                passes_into_final_third=None,
                passes_into_penalty_area=None,
                crosses_into_penalty_area=None,
                live_ball_passes=None,
                dead_ball_passes=None,
                passes_from_free_kicks=None,
                through_balls=None,
                switches=None,
                crosses=None,
                throw_ins_taken=None,
                corner_kicks=None,
                inswinging_corner_kicks=None,
                outswinging_corner_kicks=None,
                straight_corner_kicks=None,
                passes_offside=None,
                passes_blocked=None,
                tackles_won=None,
                tackles_in_defensive_third=None,
                tackles_in_middle_third=None,
                tackles_in_attacking_third=None,
                dribblers_tackled=None,
                dribbles_challenged=None,
                percent_of_dribblers_tackled=None,
                challenges_lost=None,
                shots_blocked=None,
                tackles_plus_interceptions=None,
                errors=None,
                touches_in_defensive_penalty_area=None,
                touches_in_defensive_third=None,
                touches_in_middle_third=None,
                touches_in_attacking_third=None,
                touches_in_attacking_penalty_area=None,
                live_ball_touches=None,
                successful_take_on_percentage=None,
                times_tackled_during_take_ons=None,
                tackled_during_take_on_percentage=None,
                total_carrying_distance=None,
                progressive_carrying_distance=None,
                carries_into_final_third=None,
                carries_into_penalty_area=None,
                miscontrols=None,
                dispossessed=None,
                passes_received=None,
                progressive_passes_received=None,
                second_yellow_card=None,
                fouls_committed=None,
                fouls_drawn=None,
                offsides=None,
                penalty_kicks_won=None,
                penalty_kicks_conceded=None,
                own_goals=None,
                ball_recoveries=None,
                aerials_won=None,
                aerials_lost=None,
                percentage_of_aerials_won=None,
                shots_on_target_against=None,
                post_shot_expected_goals=None,
                passes_attempted_minus_goal_kicks=None,
                throws_attempted=None,
                percentage_of_passes_that_were_launched=None,
                average_pass_length=None,
                goal_kicks_attempted=None,
                percentage_of_goal_kicks_that_were_launched=None,
                average_goal_kick_length=None,
                crosses_faced=None,
                crosses_stopped=None,
                percentage_crosses_stopped=None,
                defensive_actions_outside_penalty_area=None,
                average_distance_of_defensive_actions=None,
                three_point_attempt_rate=None,
            )]
            team_models = [TeamModel(
                identifier="a",
                name="Team A",
                location=None,
                players=player_models,
                odds=[],
                points=None,
                ladder_rank=None,
                news=[],
                social=[],
                coaches=[],
                lbw=None,
                end_dt=None,
                version=VERSION,
            )]
            team_model = create_combined_team_model(
                team_models=team_models,
                identifier="a",
                player_identity_map={},
                names=names,
                coach_names=coach_names,
                player_ffill=player_ffill,
                team_ffill=team_ffill,
                coach_ffill=coach_ffill,
            )
            player_models_2 = [PlayerModel(
                identifier="1a",
                jersey=None,
                kicks=None,
                fumbles=None,
                fumbles_lost=None,
                field_goals=None,
                field_goals_attempted=None,
                offensive_rebounds=None,
                assists=None,
                turnovers=None,
                name="Johnny Johnson",
                marks=None,
                handballs=None,
                disposals=None,
                goals=None,
                behinds=None,
                hit_outs=None,
                tackles=None,
                rebounds=None,
                insides=None,
                clearances=None,
                clangers=None,
                free_kicks_for=None,
                free_kicks_against=None,
                brownlow_votes=None,
                contested_possessions=None,
                uncontested_possessions=None,
                contested_marks=None,
                marks_inside=None,
                one_percenters=None,
                bounces=None,
                goal_assists=None,
                percentage_played=None,
                birth_date=None,
                age=None,
                species=str(Species.HUMAN),
                handicap_weight=None,
                father=None,
                starting_position=None,
                weight=None,
                birth_address=None,
                owner=None,
                seconds_played=None,
                three_point_field_goals=None,
                three_point_field_goals_attempted=None,
                free_throws=None,
                free_throws_attempted=None,
                defensive_rebounds=None,
                steals=None,
                blocks=None,
                personal_fouls=None,
                points=None,
                game_score=None,
                point_differential=None,
                version=PLAYER_VERSION,
                height=None,
                colleges=[],
                headshot=None,
                forced_fumbles=None,
                fumbles_recovered=None,
                fumbles_recovered_yards=None,
                fumbles_touchdowns=None,
                offensive_two_point_returns=None,
                offensive_fumbles_touchdowns=None,
                defensive_fumbles_touchdowns=None,
                average_gain=None,
                completion_percentage=None,
                completions=None,
                espn_quarterback_rating=None,
                interception_percentage=None,
                interceptions=None,
                long_passing=None,
                misc_yards=None,
                net_passing_yards=None,
                net_total_yards=None,
                passing_attempts=None,
                passing_big_plays=None,
                passing_first_downs=None,
                passing_fumbles=None,
                passing_fumbles_lost=None,
                passing_touchdown_percentage=None,
                passing_touchdowns=None,
                passing_yards=None,
                passing_yards_after_catch=None,
                quarterback_rating=None,
                sacks=None,
                passing_yards_at_catch=None,
                sacks_yards_lost=None,
                net_passing_attempts=None,
                total_offensive_plays=None,
                total_points=None,
                total_touchdowns=None,
                total_yards=None,
                total_yards_from_scrimmage=None,
                two_point_pass=None,
                two_point_pass_attempt=None,
                yards_per_completion=None,
                yards_per_pass_attempt=None,
                net_yards_per_pass_attempt=None,
                long_rushing=None,
                rushing_attempts=None,
                rushing_big_plays=None,
                rushing_first_downs=None,
                rushing_fumbles=None,
                rushing_fumbles_lost=None,
                rushing_touchdowns=None,
                rushing_yards=None,
                stuffs=None,
                stuff_yards_lost=None,
                two_point_rush=None,
                two_point_rush_attempts=None,
                yards_per_rush_attempt=None,
                espn_widereceiver=None,
                long_reception=None,
                receiving_big_plays=None,
                receiving_first_downs=None,
                receiving_fumbles=None,
                receiving_fumbles_lost=None,
                receiving_targets=None,
                receiving_touchdowns=None,
                receiving_yards=None,
                receiving_yards_after_catch=None,
                receiving_yards_at_catch=None,
                receptions=None,
                two_point_receptions=None,
                two_point_reception_attempts=None,
                yards_per_reception=None,
                assist_tackles=None,
                average_interception_yards=None,
                average_sack_yards=None,
                average_stuff_yards=None,
                blocked_field_goal_touchdowns=None,
                blocked_punt_touchdowns=None,
                defensive_touchdowns=None,
                hurries=None,
                kicks_blocked=None,
                long_interception=None,
                misc_touchdowns=None,
                passes_batted_down=None,
                passes_defended=None,
                quarterback_hits=None,
                sacks_assisted=None,
                sacks_unassisted=None,
                sacks_yards=None,
                safeties=None,
                solo_tackles=None,
                stuff_yards=None,
                tackles_for_loss=None,
                tackles_yards_lost=None,
                yards_allowed=None,
                points_allowed=None,
                one_point_safeties_made=None,
                missed_field_goal_return_td=None,
                blocked_punt_ez_rec_td=None,
                interception_touchdowns=None,
                interception_yards=None,
                average_kickoff_return_yards=None,
                average_kickoff_yards=None,
                extra_point_attempts=None,
                extra_point_percentage=None,
                extra_point_blocked=None,
                extra_points_blocked_percentage=None,
                extra_points_made=None,
                fair_catches=None,
                fair_catch_percentage=None,
                field_goal_attempts_max_19_yards=None,
                field_goal_attempts_max_29_yards=None,
                field_goal_attempts_max_39_yards=None,
                field_goal_attempts_max_49_yards=None,
                field_goal_attempts_max_59_yards=None,
                field_goal_attempts_max_99_yards=None,
                field_goal_attempts_above_50_yards=None,
                field_goal_attempt_yards=None,
                field_goals_blocked=None,
                field_goals_blocked_percentage=None,
                field_goals_made=None,
                field_goals_made_max_19_yards=None,
                field_goals_made_max_29_yards=None,
                field_goals_made_max_39_yards=None,
                field_goals_made_max_49_yards=None,
                field_goals_made_max_59_yards=None,
                field_goals_made_max_99_yards=None,
                field_goals_made_above_50_yards=None,
                field_goals_made_yards=None,
                field_goals_missed_yards=None,
                kickoff_out_of_bounds=None,
                kickoff_returns=None,
                kickoff_returns_touchdowns=None,
                kickoff_return_yards=None,
                kickoffs=None,
                kickoff_yards=None,
                long_field_goal_attempt=None,
                long_field_goal_made=None,
                long_kickoff=None,
                total_kicking_points=None,
                touchback_percentage=None,
                touchbacks=None,
                defensive_fumble_returns=None,
                defensive_fumble_return_yards=None,
                fumble_recoveries=None,
                fumble_recovery_yards=None,
                kick_return_fair_catches=None,
                kick_return_fair_catch_percentage=None,
                kick_return_fumbles=None,
                kick_return_fumbles_lost=None,
                kick_returns=None,
                kick_return_touchdowns=None,
                kick_return_yards=None,
                long_kick_return=None,
                long_punt_return=None,
                misc_fumble_returns=None,
                misc_fumble_return_yards=None,
                opposition_fumble_recoveries=None,
                opposition_fumble_recovery_yards=None,
                opposition_special_team_fumble_returns=None,
                opposition_special_team_fumble_return_yards=None,
                punt_return_fair_catches=None,
                punt_return_fair_catch_percentage=None,
                punt_return_fumbles=None,
                punt_return_fumbles_lost=None,
                punt_returns=None,
                punt_returns_started_inside_the_10=None,
                punt_returns_started_inside_the_20=None,
                punt_return_touchdowns=None,
                special_team_fumble_returns=None,
                yards_per_kick_return=None,
                yards_per_punt_return=None,
                yards_per_return=None,
                average_punt_return_yards=None,
                gross_average_punt_yards=None,
                long_punt=None,
                net_average_punt_yards=None,
                punts=None,
                punts_blocked=None,
                punts_blocked_percentage=None,
                punts_inside_10=None,
                punts_inside_10_percentage=None,
                punts_inside_20=None,
                punts_inside_20_percentage=None,
                punts_over_50=None,
                punt_yards=None,
                defensive_points=None,
                misc_points=None,
                return_touchdowns=None,
                total_two_point_conversions=None,
                passing_touchdowns_9_yards=None,
                passing_touchdowns_19_yards=None,
                passing_touchdowns_29_yards=None,
                passing_touchdowns_39_yards=None,
                passing_touchdowns_49_yards=None,
                passing_touchdowns_above_50_yards=None,
                receiving_touchdowns_9_yards=None,
                receiving_touchdowns_19_yards=None,
                receiving_touchdowns_29_yards=None,
                receiving_touchdowns_39_yards=None,
                punt_return_yards=None,
                receiving_touchdowns_49_yards=None,
                receiving_touchdowns_above_50_yards=None,
                rushing_touchdowns_9_yards=None,
                rushing_touchdowns_19_yards=None,
                rushing_touchdowns_29_yards=None,
                rushing_touchdowns_39_yards=None,
                rushing_touchdowns_49_yards=None,
                rushing_touchdowns_above_50_yards=None,
                penalties_in_minutes=None,
                even_strength_goals=None,
                power_play_goals=None,
                short_handed_goals=None,
                game_winning_goals=None,
                even_strength_assists=None,
                power_play_assists=None,
                short_handed_assists=None,
                shots_on_goal=None,
                shooting_percentage=None,
                shifts=None,
                time_on_ice=None,
                decision=None,
                goals_against=None,
                shots_against=None,
                saves=None,
                save_percentage=None,
                shutouts=None,
                individual_corsi_for_events=None,
                on_shot_ice_for_events=None,
                on_shot_ice_against_events=None,
                corsi_for_percentage=None,
                relative_corsi_for_percentage=None,
                offensive_zone_starts=None,
                defensive_zone_starts=None,
                offensive_zone_start_percentage=None,
                hits=None,
                true_shooting_percentage=None,
                at_bats=None,
                runs_scored=None,
                runs_batted_in=None,
                bases_on_balls=None,
                strikeouts=None,
                plate_appearances=None,
                hits_at_bats=None,
                obp=None,
                slg=None,
                ops=None,
                pitches=None,
                strikes=None,
                win_probability_added=None,
                average_leverage_index=None,
                wpa_plus=None,
                wpa_minus=None,
                cwpa=None,
                acli=None,
                re24=None,
                putouts=None,
                innings_pitched=None,
                earned_runs=None,
                home_runs=None,
                era=None,
                batters_faced=None,
                strikes_by_contact=None,
                strikes_swinging=None,
                strikes_looking=None,
                ground_balls=None,
                fly_balls=None,
                line_drives=None,
                inherited_runners=None,
                inherited_scores=None,
                effective_field_goal_percentage=None,
                penalty_kicks_made=None,
                penalty_kicks_attempted=None,
                shots_total=None,
                shots_on_target=None,
                yellow_cards=None,
                red_cards=None,
                touches=None,
                expected_goals=None,
                non_penalty_expected_goals=None,
                expected_assisted_goals=None,
                shot_creating_actions=None,
                goal_creating_actions=None,
                passes_completed=None,
                passes_attempted=None,
                pass_completion=None,
                progressive_passes=None,
                carries=None,
                progressive_carries=None,
                take_ons_attempted=None,
                successful_take_ons=None,
                total_passing_distance=None,
                progressive_passing_distance=None,
                passes_completed_short=None,
                passes_attempted_short=None,
                pass_completion_short=None,
                passes_completed_medium=None,
                passes_attempted_medium=None,
                pass_completion_medium=None,
                passes_completed_long=None,
                passes_attempted_long=None,
                pass_completion_long=None,
                expected_assists=None,
                key_passes=None,
                passes_into_final_third=None,
                passes_into_penalty_area=None,
                crosses_into_penalty_area=None,
                live_ball_passes=None,
                dead_ball_passes=None,
                passes_from_free_kicks=None,
                through_balls=None,
                switches=None,
                crosses=None,
                throw_ins_taken=None,
                corner_kicks=None,
                inswinging_corner_kicks=None,
                outswinging_corner_kicks=None,
                straight_corner_kicks=None,
                passes_offside=None,
                passes_blocked=None,
                tackles_won=None,
                tackles_in_defensive_third=None,
                tackles_in_middle_third=None,
                tackles_in_attacking_third=None,
                dribblers_tackled=None,
                dribbles_challenged=None,
                percent_of_dribblers_tackled=None,
                challenges_lost=None,
                shots_blocked=None,
                tackles_plus_interceptions=None,
                errors=None,
                touches_in_defensive_penalty_area=None,
                touches_in_defensive_third=None,
                touches_in_middle_third=None,
                touches_in_attacking_third=None,
                touches_in_attacking_penalty_area=None,
                live_ball_touches=None,
                successful_take_on_percentage=None,
                times_tackled_during_take_ons=None,
                tackled_during_take_on_percentage=None,
                total_carrying_distance=None,
                progressive_carrying_distance=None,
                carries_into_final_third=None,
                carries_into_penalty_area=None,
                miscontrols=None,
                dispossessed=None,
                passes_received=None,
                progressive_passes_received=None,
                second_yellow_card=None,
                fouls_committed=None,
                fouls_drawn=None,
                offsides=None,
                penalty_kicks_won=None,
                penalty_kicks_conceded=None,
                own_goals=None,
                ball_recoveries=None,
                aerials_won=None,
                aerials_lost=None,
                percentage_of_aerials_won=None,
                shots_on_target_against=None,
                post_shot_expected_goals=None,
                passes_attempted_minus_goal_kicks=None,
                throws_attempted=None,
                percentage_of_passes_that_were_launched=None,
                average_pass_length=None,
                goal_kicks_attempted=None,
                percentage_of_goal_kicks_that_were_launched=None,
                average_goal_kick_length=None,
                crosses_faced=None,
                crosses_stopped=None,
                percentage_crosses_stopped=None,
                defensive_actions_outside_penalty_area=None,
                average_distance_of_defensive_actions=None,
                three_point_attempt_rate=None,
            )]
            team_models_2 = [TeamModel(
                identifier="a",
                name="Team A",
                location=None,
                players=player_models_2,
                odds=[],
                points=None,
                ladder_rank=None,
                news=[],
                social=[],
                coaches=[],
                lbw=None,
                end_dt=None,
                version=VERSION,
            )]
            team_model_2 = create_combined_team_model(
                team_models=team_models_2,
                identifier="a",
                player_identity_map={},
                names=names,
                coach_names=coach_names,
                player_ffill=player_ffill,
                team_ffill=team_ffill,
                coach_ffill=coach_ffill,
            )
            self.assertEqual(team_model.players[0].identifier, team_model_2.players[0].identifier)

    def test_names_resolve_with_surname_firstname(self):
        with requests_mock.Mocker() as m:
            names = {}
            coach_names = {}
            player_ffill = {}
            team_ffill = {}
            coach_ffill = {}
            player_models = [PlayerModel(
                identifier="1",
                jersey=None,
                kicks=None,
                fumbles=None,
                fumbles_lost=None,
                field_goals=None,
                field_goals_attempted=None,
                offensive_rebounds=None,
                assists=None,
                turnovers=None,
                name="Johnny Johnson",
                marks=None,
                handballs=None,
                disposals=None,
                goals=None,
                behinds=None,
                hit_outs=None,
                tackles=None,
                rebounds=None,
                insides=None,
                clearances=None,
                clangers=None,
                free_kicks_for=None,
                free_kicks_against=None,
                brownlow_votes=None,
                contested_possessions=None,
                uncontested_possessions=None,
                contested_marks=None,
                marks_inside=None,
                one_percenters=None,
                bounces=None,
                goal_assists=None,
                percentage_played=None,
                birth_date=None,
                age=None,
                species=str(Species.HUMAN),
                handicap_weight=None,
                father=None,
                starting_position=None,
                weight=None,
                birth_address=None,
                owner=None,
                seconds_played=None,
                three_point_field_goals=None,
                three_point_field_goals_attempted=None,
                free_throws=None,
                free_throws_attempted=None,
                defensive_rebounds=None,
                steals=None,
                blocks=None,
                personal_fouls=None,
                points=None,
                game_score=None,
                point_differential=None,
                version=PLAYER_VERSION,
                height=None,
                colleges=[],
                headshot=None,
                forced_fumbles=None,
                fumbles_recovered=None,
                fumbles_recovered_yards=None,
                fumbles_touchdowns=None,
                offensive_two_point_returns=None,
                offensive_fumbles_touchdowns=None,
                defensive_fumbles_touchdowns=None,
                average_gain=None,
                completion_percentage=None,
                completions=None,
                espn_quarterback_rating=None,
                interception_percentage=None,
                interceptions=None,
                long_passing=None,
                misc_yards=None,
                net_passing_yards=None,
                net_total_yards=None,
                passing_attempts=None,
                passing_big_plays=None,
                passing_first_downs=None,
                passing_fumbles=None,
                passing_fumbles_lost=None,
                passing_touchdown_percentage=None,
                passing_touchdowns=None,
                passing_yards=None,
                passing_yards_after_catch=None,
                quarterback_rating=None,
                sacks=None,
                passing_yards_at_catch=None,
                sacks_yards_lost=None,
                net_passing_attempts=None,
                total_offensive_plays=None,
                total_points=None,
                total_touchdowns=None,
                total_yards=None,
                total_yards_from_scrimmage=None,
                two_point_pass=None,
                two_point_pass_attempt=None,
                yards_per_completion=None,
                yards_per_pass_attempt=None,
                net_yards_per_pass_attempt=None,
                long_rushing=None,
                rushing_attempts=None,
                rushing_big_plays=None,
                rushing_first_downs=None,
                rushing_fumbles=None,
                rushing_fumbles_lost=None,
                rushing_touchdowns=None,
                rushing_yards=None,
                stuffs=None,
                stuff_yards_lost=None,
                two_point_rush=None,
                two_point_rush_attempts=None,
                yards_per_rush_attempt=None,
                espn_widereceiver=None,
                long_reception=None,
                receiving_big_plays=None,
                receiving_first_downs=None,
                receiving_fumbles=None,
                receiving_fumbles_lost=None,
                receiving_targets=None,
                receiving_touchdowns=None,
                receiving_yards=None,
                receiving_yards_after_catch=None,
                receiving_yards_at_catch=None,
                receptions=None,
                two_point_receptions=None,
                two_point_reception_attempts=None,
                yards_per_reception=None,
                assist_tackles=None,
                average_interception_yards=None,
                average_sack_yards=None,
                average_stuff_yards=None,
                blocked_field_goal_touchdowns=None,
                blocked_punt_touchdowns=None,
                defensive_touchdowns=None,
                hurries=None,
                kicks_blocked=None,
                long_interception=None,
                misc_touchdowns=None,
                passes_batted_down=None,
                passes_defended=None,
                quarterback_hits=None,
                sacks_assisted=None,
                sacks_unassisted=None,
                sacks_yards=None,
                safeties=None,
                solo_tackles=None,
                stuff_yards=None,
                tackles_for_loss=None,
                tackles_yards_lost=None,
                yards_allowed=None,
                points_allowed=None,
                one_point_safeties_made=None,
                missed_field_goal_return_td=None,
                blocked_punt_ez_rec_td=None,
                interception_touchdowns=None,
                interception_yards=None,
                average_kickoff_return_yards=None,
                average_kickoff_yards=None,
                extra_point_attempts=None,
                extra_point_percentage=None,
                extra_point_blocked=None,
                extra_points_blocked_percentage=None,
                extra_points_made=None,
                fair_catches=None,
                fair_catch_percentage=None,
                field_goal_attempts_max_19_yards=None,
                field_goal_attempts_max_29_yards=None,
                field_goal_attempts_max_39_yards=None,
                field_goal_attempts_max_49_yards=None,
                field_goal_attempts_max_59_yards=None,
                field_goal_attempts_max_99_yards=None,
                field_goal_attempts_above_50_yards=None,
                field_goal_attempt_yards=None,
                field_goals_blocked=None,
                field_goals_blocked_percentage=None,
                field_goals_made=None,
                field_goals_made_max_19_yards=None,
                field_goals_made_max_29_yards=None,
                field_goals_made_max_39_yards=None,
                field_goals_made_max_49_yards=None,
                field_goals_made_max_59_yards=None,
                field_goals_made_max_99_yards=None,
                field_goals_made_above_50_yards=None,
                field_goals_made_yards=None,
                field_goals_missed_yards=None,
                kickoff_out_of_bounds=None,
                kickoff_returns=None,
                kickoff_returns_touchdowns=None,
                kickoff_return_yards=None,
                kickoffs=None,
                kickoff_yards=None,
                long_field_goal_attempt=None,
                long_field_goal_made=None,
                long_kickoff=None,
                total_kicking_points=None,
                touchback_percentage=None,
                touchbacks=None,
                defensive_fumble_returns=None,
                defensive_fumble_return_yards=None,
                fumble_recoveries=None,
                fumble_recovery_yards=None,
                kick_return_fair_catches=None,
                kick_return_fair_catch_percentage=None,
                kick_return_fumbles=None,
                kick_return_fumbles_lost=None,
                kick_returns=None,
                kick_return_touchdowns=None,
                kick_return_yards=None,
                long_kick_return=None,
                long_punt_return=None,
                misc_fumble_returns=None,
                misc_fumble_return_yards=None,
                opposition_fumble_recoveries=None,
                opposition_fumble_recovery_yards=None,
                opposition_special_team_fumble_returns=None,
                opposition_special_team_fumble_return_yards=None,
                punt_return_fair_catches=None,
                punt_return_fair_catch_percentage=None,
                punt_return_fumbles=None,
                punt_return_fumbles_lost=None,
                punt_returns=None,
                punt_returns_started_inside_the_10=None,
                punt_returns_started_inside_the_20=None,
                punt_return_touchdowns=None,
                special_team_fumble_returns=None,
                yards_per_kick_return=None,
                yards_per_punt_return=None,
                yards_per_return=None,
                average_punt_return_yards=None,
                gross_average_punt_yards=None,
                long_punt=None,
                net_average_punt_yards=None,
                punts=None,
                punts_blocked=None,
                punts_blocked_percentage=None,
                punts_inside_10=None,
                punts_inside_10_percentage=None,
                punts_inside_20=None,
                punts_inside_20_percentage=None,
                punts_over_50=None,
                punt_yards=None,
                defensive_points=None,
                misc_points=None,
                return_touchdowns=None,
                total_two_point_conversions=None,
                passing_touchdowns_9_yards=None,
                passing_touchdowns_19_yards=None,
                passing_touchdowns_29_yards=None,
                passing_touchdowns_39_yards=None,
                passing_touchdowns_49_yards=None,
                passing_touchdowns_above_50_yards=None,
                receiving_touchdowns_9_yards=None,
                receiving_touchdowns_19_yards=None,
                receiving_touchdowns_29_yards=None,
                receiving_touchdowns_39_yards=None,
                punt_return_yards=None,
                receiving_touchdowns_49_yards=None,
                receiving_touchdowns_above_50_yards=None,
                rushing_touchdowns_9_yards=None,
                rushing_touchdowns_19_yards=None,
                rushing_touchdowns_29_yards=None,
                rushing_touchdowns_39_yards=None,
                rushing_touchdowns_49_yards=None,
                rushing_touchdowns_above_50_yards=None,
                penalties_in_minutes=None,
                even_strength_goals=None,
                power_play_goals=None,
                short_handed_goals=None,
                game_winning_goals=None,
                even_strength_assists=None,
                power_play_assists=None,
                short_handed_assists=None,
                shots_on_goal=None,
                shooting_percentage=None,
                shifts=None,
                time_on_ice=None,
                decision=None,
                goals_against=None,
                shots_against=None,
                saves=None,
                save_percentage=None,
                shutouts=None,
                individual_corsi_for_events=None,
                on_shot_ice_for_events=None,
                on_shot_ice_against_events=None,
                corsi_for_percentage=None,
                relative_corsi_for_percentage=None,
                offensive_zone_starts=None,
                defensive_zone_starts=None,
                offensive_zone_start_percentage=None,
                hits=None,
                true_shooting_percentage=None,
                at_bats=None,
                runs_scored=None,
                runs_batted_in=None,
                bases_on_balls=None,
                strikeouts=None,
                plate_appearances=None,
                hits_at_bats=None,
                obp=None,
                slg=None,
                ops=None,
                pitches=None,
                strikes=None,
                win_probability_added=None,
                average_leverage_index=None,
                wpa_plus=None,
                wpa_minus=None,
                cwpa=None,
                acli=None,
                re24=None,
                putouts=None,
                innings_pitched=None,
                earned_runs=None,
                home_runs=None,
                era=None,
                batters_faced=None,
                strikes_by_contact=None,
                strikes_swinging=None,
                strikes_looking=None,
                ground_balls=None,
                fly_balls=None,
                line_drives=None,
                inherited_runners=None,
                inherited_scores=None,
                effective_field_goal_percentage=None,
                penalty_kicks_made=None,
                penalty_kicks_attempted=None,
                shots_total=None,
                shots_on_target=None,
                yellow_cards=None,
                red_cards=None,
                touches=None,
                expected_goals=None,
                non_penalty_expected_goals=None,
                expected_assisted_goals=None,
                shot_creating_actions=None,
                goal_creating_actions=None,
                passes_completed=None,
                passes_attempted=None,
                pass_completion=None,
                progressive_passes=None,
                carries=None,
                progressive_carries=None,
                take_ons_attempted=None,
                successful_take_ons=None,
                total_passing_distance=None,
                progressive_passing_distance=None,
                passes_completed_short=None,
                passes_attempted_short=None,
                pass_completion_short=None,
                passes_completed_medium=None,
                passes_attempted_medium=None,
                pass_completion_medium=None,
                passes_completed_long=None,
                passes_attempted_long=None,
                pass_completion_long=None,
                expected_assists=None,
                key_passes=None,
                passes_into_final_third=None,
                passes_into_penalty_area=None,
                crosses_into_penalty_area=None,
                live_ball_passes=None,
                dead_ball_passes=None,
                passes_from_free_kicks=None,
                through_balls=None,
                switches=None,
                crosses=None,
                throw_ins_taken=None,
                corner_kicks=None,
                inswinging_corner_kicks=None,
                outswinging_corner_kicks=None,
                straight_corner_kicks=None,
                passes_offside=None,
                passes_blocked=None,
                tackles_won=None,
                tackles_in_defensive_third=None,
                tackles_in_middle_third=None,
                tackles_in_attacking_third=None,
                dribblers_tackled=None,
                dribbles_challenged=None,
                percent_of_dribblers_tackled=None,
                challenges_lost=None,
                shots_blocked=None,
                tackles_plus_interceptions=None,
                errors=None,
                touches_in_defensive_penalty_area=None,
                touches_in_defensive_third=None,
                touches_in_middle_third=None,
                touches_in_attacking_third=None,
                touches_in_attacking_penalty_area=None,
                live_ball_touches=None,
                successful_take_on_percentage=None,
                times_tackled_during_take_ons=None,
                tackled_during_take_on_percentage=None,
                total_carrying_distance=None,
                progressive_carrying_distance=None,
                carries_into_final_third=None,
                carries_into_penalty_area=None,
                miscontrols=None,
                dispossessed=None,
                passes_received=None,
                progressive_passes_received=None,
                second_yellow_card=None,
                fouls_committed=None,
                fouls_drawn=None,
                offsides=None,
                penalty_kicks_won=None,
                penalty_kicks_conceded=None,
                own_goals=None,
                ball_recoveries=None,
                aerials_won=None,
                aerials_lost=None,
                percentage_of_aerials_won=None,
                shots_on_target_against=None,
                post_shot_expected_goals=None,
                passes_attempted_minus_goal_kicks=None,
                throws_attempted=None,
                percentage_of_passes_that_were_launched=None,
                average_pass_length=None,
                goal_kicks_attempted=None,
                percentage_of_goal_kicks_that_were_launched=None,
                average_goal_kick_length=None,
                crosses_faced=None,
                crosses_stopped=None,
                percentage_crosses_stopped=None,
                defensive_actions_outside_penalty_area=None,
                average_distance_of_defensive_actions=None,
                three_point_attempt_rate=None,
            )]
            team_models = [TeamModel(
                identifier="a",
                name="Team A",
                location=None,
                players=player_models,
                odds=[],
                points=None,
                ladder_rank=None,
                news=[],
                social=[],
                coaches=[],
                lbw=None,
                end_dt=None,
                version=VERSION,
            )]
            team_model = create_combined_team_model(
                team_models=team_models,
                identifier="a",
                player_identity_map={},
                names=names,
                coach_names=coach_names,
                player_ffill=player_ffill,
                team_ffill=team_ffill,
                coach_ffill=coach_ffill,
            )
            player_models_2 = [PlayerModel(
                identifier="1a",
                jersey=None,
                kicks=None,
                fumbles=None,
                fumbles_lost=None,
                field_goals=None,
                field_goals_attempted=None,
                offensive_rebounds=None,
                assists=None,
                turnovers=None,
                name="Johnson, Johnny",
                marks=None,
                handballs=None,
                disposals=None,
                goals=None,
                behinds=None,
                hit_outs=None,
                tackles=None,
                rebounds=None,
                insides=None,
                clearances=None,
                clangers=None,
                free_kicks_for=None,
                free_kicks_against=None,
                brownlow_votes=None,
                contested_possessions=None,
                uncontested_possessions=None,
                contested_marks=None,
                marks_inside=None,
                one_percenters=None,
                bounces=None,
                goal_assists=None,
                percentage_played=None,
                birth_date=None,
                age=None,
                species=str(Species.HUMAN),
                handicap_weight=None,
                father=None,
                starting_position=None,
                weight=None,
                birth_address=None,
                owner=None,
                seconds_played=None,
                three_point_field_goals=None,
                three_point_field_goals_attempted=None,
                free_throws=None,
                free_throws_attempted=None,
                defensive_rebounds=None,
                steals=None,
                blocks=None,
                personal_fouls=None,
                points=None,
                game_score=None,
                point_differential=None,
                version=PLAYER_VERSION,
                height=None,
                colleges=[],
                headshot=None,
                forced_fumbles=None,
                fumbles_recovered=None,
                fumbles_recovered_yards=None,
                fumbles_touchdowns=None,
                offensive_two_point_returns=None,
                offensive_fumbles_touchdowns=None,
                defensive_fumbles_touchdowns=None,
                average_gain=None,
                completion_percentage=None,
                completions=None,
                espn_quarterback_rating=None,
                interception_percentage=None,
                interceptions=None,
                long_passing=None,
                misc_yards=None,
                net_passing_yards=None,
                net_total_yards=None,
                passing_attempts=None,
                passing_big_plays=None,
                passing_first_downs=None,
                passing_fumbles=None,
                passing_fumbles_lost=None,
                passing_touchdown_percentage=None,
                passing_touchdowns=None,
                passing_yards=None,
                passing_yards_after_catch=None,
                quarterback_rating=None,
                sacks=None,
                passing_yards_at_catch=None,
                sacks_yards_lost=None,
                net_passing_attempts=None,
                total_offensive_plays=None,
                total_points=None,
                total_touchdowns=None,
                total_yards=None,
                total_yards_from_scrimmage=None,
                two_point_pass=None,
                two_point_pass_attempt=None,
                yards_per_completion=None,
                yards_per_pass_attempt=None,
                net_yards_per_pass_attempt=None,
                long_rushing=None,
                rushing_attempts=None,
                rushing_big_plays=None,
                rushing_first_downs=None,
                rushing_fumbles=None,
                rushing_fumbles_lost=None,
                rushing_touchdowns=None,
                rushing_yards=None,
                stuffs=None,
                stuff_yards_lost=None,
                two_point_rush=None,
                two_point_rush_attempts=None,
                yards_per_rush_attempt=None,
                espn_widereceiver=None,
                long_reception=None,
                receiving_big_plays=None,
                receiving_first_downs=None,
                receiving_fumbles=None,
                receiving_fumbles_lost=None,
                receiving_targets=None,
                receiving_touchdowns=None,
                receiving_yards=None,
                receiving_yards_after_catch=None,
                receiving_yards_at_catch=None,
                receptions=None,
                two_point_receptions=None,
                two_point_reception_attempts=None,
                yards_per_reception=None,
                assist_tackles=None,
                average_interception_yards=None,
                average_sack_yards=None,
                average_stuff_yards=None,
                blocked_field_goal_touchdowns=None,
                blocked_punt_touchdowns=None,
                defensive_touchdowns=None,
                hurries=None,
                kicks_blocked=None,
                long_interception=None,
                misc_touchdowns=None,
                passes_batted_down=None,
                passes_defended=None,
                quarterback_hits=None,
                sacks_assisted=None,
                sacks_unassisted=None,
                sacks_yards=None,
                safeties=None,
                solo_tackles=None,
                stuff_yards=None,
                tackles_for_loss=None,
                tackles_yards_lost=None,
                yards_allowed=None,
                points_allowed=None,
                one_point_safeties_made=None,
                missed_field_goal_return_td=None,
                blocked_punt_ez_rec_td=None,
                interception_touchdowns=None,
                interception_yards=None,
                average_kickoff_return_yards=None,
                average_kickoff_yards=None,
                extra_point_attempts=None,
                extra_point_percentage=None,
                extra_point_blocked=None,
                extra_points_blocked_percentage=None,
                extra_points_made=None,
                fair_catches=None,
                fair_catch_percentage=None,
                field_goal_attempts_max_19_yards=None,
                field_goal_attempts_max_29_yards=None,
                field_goal_attempts_max_39_yards=None,
                field_goal_attempts_max_49_yards=None,
                field_goal_attempts_max_59_yards=None,
                field_goal_attempts_max_99_yards=None,
                field_goal_attempts_above_50_yards=None,
                field_goal_attempt_yards=None,
                field_goals_blocked=None,
                field_goals_blocked_percentage=None,
                field_goals_made=None,
                field_goals_made_max_19_yards=None,
                field_goals_made_max_29_yards=None,
                field_goals_made_max_39_yards=None,
                field_goals_made_max_49_yards=None,
                field_goals_made_max_59_yards=None,
                field_goals_made_max_99_yards=None,
                field_goals_made_above_50_yards=None,
                field_goals_made_yards=None,
                field_goals_missed_yards=None,
                kickoff_out_of_bounds=None,
                kickoff_returns=None,
                kickoff_returns_touchdowns=None,
                kickoff_return_yards=None,
                kickoffs=None,
                kickoff_yards=None,
                long_field_goal_attempt=None,
                long_field_goal_made=None,
                long_kickoff=None,
                total_kicking_points=None,
                touchback_percentage=None,
                touchbacks=None,
                defensive_fumble_returns=None,
                defensive_fumble_return_yards=None,
                fumble_recoveries=None,
                fumble_recovery_yards=None,
                kick_return_fair_catches=None,
                kick_return_fair_catch_percentage=None,
                kick_return_fumbles=None,
                kick_return_fumbles_lost=None,
                kick_returns=None,
                kick_return_touchdowns=None,
                kick_return_yards=None,
                long_kick_return=None,
                long_punt_return=None,
                misc_fumble_returns=None,
                misc_fumble_return_yards=None,
                opposition_fumble_recoveries=None,
                opposition_fumble_recovery_yards=None,
                opposition_special_team_fumble_returns=None,
                opposition_special_team_fumble_return_yards=None,
                punt_return_fair_catches=None,
                punt_return_fair_catch_percentage=None,
                punt_return_fumbles=None,
                punt_return_fumbles_lost=None,
                punt_returns=None,
                punt_returns_started_inside_the_10=None,
                punt_returns_started_inside_the_20=None,
                punt_return_touchdowns=None,
                special_team_fumble_returns=None,
                yards_per_kick_return=None,
                yards_per_punt_return=None,
                yards_per_return=None,
                average_punt_return_yards=None,
                gross_average_punt_yards=None,
                long_punt=None,
                net_average_punt_yards=None,
                punts=None,
                punts_blocked=None,
                punts_blocked_percentage=None,
                punts_inside_10=None,
                punts_inside_10_percentage=None,
                punts_inside_20=None,
                punts_inside_20_percentage=None,
                punts_over_50=None,
                punt_yards=None,
                defensive_points=None,
                misc_points=None,
                return_touchdowns=None,
                total_two_point_conversions=None,
                passing_touchdowns_9_yards=None,
                passing_touchdowns_19_yards=None,
                passing_touchdowns_29_yards=None,
                passing_touchdowns_39_yards=None,
                passing_touchdowns_49_yards=None,
                passing_touchdowns_above_50_yards=None,
                receiving_touchdowns_9_yards=None,
                receiving_touchdowns_19_yards=None,
                receiving_touchdowns_29_yards=None,
                receiving_touchdowns_39_yards=None,
                punt_return_yards=None,
                receiving_touchdowns_49_yards=None,
                receiving_touchdowns_above_50_yards=None,
                rushing_touchdowns_9_yards=None,
                rushing_touchdowns_19_yards=None,
                rushing_touchdowns_29_yards=None,
                rushing_touchdowns_39_yards=None,
                rushing_touchdowns_49_yards=None,
                rushing_touchdowns_above_50_yards=None,
                penalties_in_minutes=None,
                even_strength_goals=None,
                power_play_goals=None,
                short_handed_goals=None,
                game_winning_goals=None,
                even_strength_assists=None,
                power_play_assists=None,
                short_handed_assists=None,
                shots_on_goal=None,
                shooting_percentage=None,
                shifts=None,
                time_on_ice=None,
                decision=None,
                goals_against=None,
                shots_against=None,
                saves=None,
                save_percentage=None,
                shutouts=None,
                individual_corsi_for_events=None,
                on_shot_ice_for_events=None,
                on_shot_ice_against_events=None,
                corsi_for_percentage=None,
                relative_corsi_for_percentage=None,
                offensive_zone_starts=None,
                defensive_zone_starts=None,
                offensive_zone_start_percentage=None,
                hits=None,
                true_shooting_percentage=None,
                at_bats=None,
                runs_scored=None,
                runs_batted_in=None,
                bases_on_balls=None,
                strikeouts=None,
                plate_appearances=None,
                hits_at_bats=None,
                obp=None,
                slg=None,
                ops=None,
                pitches=None,
                strikes=None,
                win_probability_added=None,
                average_leverage_index=None,
                wpa_plus=None,
                wpa_minus=None,
                cwpa=None,
                acli=None,
                re24=None,
                putouts=None,
                innings_pitched=None,
                earned_runs=None,
                home_runs=None,
                era=None,
                batters_faced=None,
                strikes_by_contact=None,
                strikes_swinging=None,
                strikes_looking=None,
                ground_balls=None,
                fly_balls=None,
                line_drives=None,
                inherited_runners=None,
                inherited_scores=None,
                effective_field_goal_percentage=None,
                penalty_kicks_made=None,
                penalty_kicks_attempted=None,
                shots_total=None,
                shots_on_target=None,
                yellow_cards=None,
                red_cards=None,
                touches=None,
                expected_goals=None,
                non_penalty_expected_goals=None,
                expected_assisted_goals=None,
                shot_creating_actions=None,
                goal_creating_actions=None,
                passes_completed=None,
                passes_attempted=None,
                pass_completion=None,
                progressive_passes=None,
                carries=None,
                progressive_carries=None,
                take_ons_attempted=None,
                successful_take_ons=None,
                total_passing_distance=None,
                progressive_passing_distance=None,
                passes_completed_short=None,
                passes_attempted_short=None,
                pass_completion_short=None,
                passes_completed_medium=None,
                passes_attempted_medium=None,
                pass_completion_medium=None,
                passes_completed_long=None,
                passes_attempted_long=None,
                pass_completion_long=None,
                expected_assists=None,
                key_passes=None,
                passes_into_final_third=None,
                passes_into_penalty_area=None,
                crosses_into_penalty_area=None,
                live_ball_passes=None,
                dead_ball_passes=None,
                passes_from_free_kicks=None,
                through_balls=None,
                switches=None,
                crosses=None,
                throw_ins_taken=None,
                corner_kicks=None,
                inswinging_corner_kicks=None,
                outswinging_corner_kicks=None,
                straight_corner_kicks=None,
                passes_offside=None,
                passes_blocked=None,
                tackles_won=None,
                tackles_in_defensive_third=None,
                tackles_in_middle_third=None,
                tackles_in_attacking_third=None,
                dribblers_tackled=None,
                dribbles_challenged=None,
                percent_of_dribblers_tackled=None,
                challenges_lost=None,
                shots_blocked=None,
                tackles_plus_interceptions=None,
                errors=None,
                touches_in_defensive_penalty_area=None,
                touches_in_defensive_third=None,
                touches_in_middle_third=None,
                touches_in_attacking_third=None,
                touches_in_attacking_penalty_area=None,
                live_ball_touches=None,
                successful_take_on_percentage=None,
                times_tackled_during_take_ons=None,
                tackled_during_take_on_percentage=None,
                total_carrying_distance=None,
                progressive_carrying_distance=None,
                carries_into_final_third=None,
                carries_into_penalty_area=None,
                miscontrols=None,
                dispossessed=None,
                passes_received=None,
                progressive_passes_received=None,
                second_yellow_card=None,
                fouls_committed=None,
                fouls_drawn=None,
                offsides=None,
                penalty_kicks_won=None,
                penalty_kicks_conceded=None,
                own_goals=None,
                ball_recoveries=None,
                aerials_won=None,
                aerials_lost=None,
                percentage_of_aerials_won=None,
                shots_on_target_against=None,
                post_shot_expected_goals=None,
                passes_attempted_minus_goal_kicks=None,
                throws_attempted=None,
                percentage_of_passes_that_were_launched=None,
                average_pass_length=None,
                goal_kicks_attempted=None,
                percentage_of_goal_kicks_that_were_launched=None,
                average_goal_kick_length=None,
                crosses_faced=None,
                crosses_stopped=None,
                percentage_crosses_stopped=None,
                defensive_actions_outside_penalty_area=None,
                average_distance_of_defensive_actions=None,
                three_point_attempt_rate=None,
            )]
            team_models_2 = [TeamModel(
                identifier="a",
                name="Team A",
                location=None,
                players=player_models_2,
                odds=[],
                points=None,
                ladder_rank=None,
                news=[],
                social=[],
                coaches=[],
                lbw=None,
                end_dt=None,
                version=VERSION,
            )]
            team_model_2 = create_combined_team_model(
                team_models=team_models_2,
                identifier="a",
                player_identity_map={},
                names=names,
                coach_names=coach_names,
                player_ffill=player_ffill,
                team_ffill=team_ffill,
                coach_ffill=coach_ffill,
            )
            self.assertEqual(team_model.players[0].identifier, team_model_2.players[0].identifier)

    def test_coach_ffill(self):
        names = {}
        coach_names = {}
        player_ffill = {}
        team_ffill = {}
        coach_ffill = {}
        coach_model = CoachModel(
            identifier="c",
            name="Wayne Johnson",
            birth_date=None,
            age=None,
            version=COACH_VERSION,
        )
        team_models = [TeamModel(
            identifier="a",
            name="Team A",
            location=None,
            players=[],
            odds=[],
            points=None,
            ladder_rank=None,
            news=[],
            social=[],
            coaches=[coach_model],
            lbw=None,
            end_dt=None,
            version=VERSION,
        )]
        team_model = create_combined_team_model(
            team_models=team_models,
            identifier="a",
            player_identity_map={},
            names=names,
            coach_names=coach_names,
            player_ffill=player_ffill,
            team_ffill=team_ffill,
            coach_ffill=coach_ffill,
        )
        next_team_models = [TeamModel(
            identifier="a",
            name="Team A",
            location=None,
            players=[],
            odds=[],
            points=None,
            ladder_rank=None,
            news=[],
            social=[],
            coaches=[],
            lbw=None,
            end_dt=None,
            version=VERSION,
        )]
        next_team_model = create_combined_team_model(
            team_models=next_team_models,
            identifier="a",
            player_identity_map={},
            names=names,
            coach_names=coach_names,
            player_ffill=player_ffill,
            team_ffill=team_ffill,
            coach_ffill=coach_ffill,
        )
        self.assertTrue(bool(next_team_model.coaches))
        self.assertEqual(team_model.coaches[0].identifier, next_team_model.coaches[0].identifier)



"""AFL AFL game model."""

# pylint: disable=too-many-statements,protected-access,too-many-arguments,bare-except,duplicate-code,too-many-locals
import datetime
import re

import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup, Tag
from playwright.sync_api import Playwright

from ....playwright import ensure_install
from ...game_model import GameModel
from ...league import League
from ...team_model import VERSION
from ...venue_model import VERSION as VENUE_VERSION
from ..position import Position, position_from_str
from .afl_afl_team_model import create_afl_afl_team_model
from .afl_afl_venue_model import create_afl_afl_venue_model


def parse_players_v1(div: Tag) -> list[list[tuple[str, str, str, str, Position]]]:
    """Parses the players from a v1 format."""
    teams_players: list[list[tuple[str, str, str, str, Position]]] = [[], []]
    for div_positions_row in div.find_all(
        "div", {"class": re.compile(".*team-lineups__positions-row.*")}
    ):
        row_players: list[list[tuple[str, str, str, str, Position]]] = [[], []]
        for count, div_team_players in enumerate(
            div_positions_row.find_all(
                "div",
                {"class": re.compile(".*team-lineups__positions-players-container.*")},
            )
        ):
            team_players: list[tuple[str, str, str, str, Position]] = []
            position = None
            for span in div_team_players.find_all(
                "span", {"class": re.compile(".*team-lineups__position-meta-label.*")}
            ):
                position = position_from_str(span.get_text().strip())
            if position is None:
                raise ValueError("position is null")
            for a in div_team_players.find_all(
                "a", {"class": re.compile(".*js-player-profile-link.*")}
            ):
                player_id = "afl:" + a.get("data-player-id")
                first_name = a.get("data-first-name")
                second_name = a.get("data-surname")
                player_number = None
                for span_player_number in a.find_all(
                    "span",
                    {"class": re.compile(".*team-lineups__player-number.*")},
                ):
                    player_number = (
                        span_player_number.get_text()
                        .strip()
                        .replace("[", "")
                        .replace("]", "")
                    )
                if player_number is None:
                    raise ValueError("player_number is null")
                team_players.append(
                    (player_id, player_number, first_name, second_name, position)
                )
            row_players[count].extend(team_players)
        for count, row_players_list in enumerate(row_players):
            teams_players[count].extend(row_players_list)
    return teams_players


def _extract_odds(soup: BeautifulSoup) -> list[float]:
    odds = []
    for span in soup.find_all("span", {"class": "betting-button__money"}):
        odds.append(float(span.get_text().strip().replace("$", "")))
    return odds


def _extract_dt(soup: BeautifulSoup) -> datetime.datetime:
    for div in soup.find_all("div", {"class": re.compile(".*js-match-start-time.*")}):
        start_time = div.get("data-start-time")
        return datetime.datetime.fromisoformat(start_time)
    raise ValueError("Unable to find datetime")


def _extract_players(
    soup: BeautifulSoup, players: list[list[tuple[str, str, str, str, Position]]]
) -> list[list[tuple[str, str, str, str, Position]]]:
    def is_players_empty() -> bool:
        nonlocal players
        if not players:
            return True
        for team in players:
            if not team:
                return True
        return False

    if not is_players_empty():
        return players

    for div in soup.find_all("div", {"class": re.compile(".*js-match-list-item.*")}):
        players = parse_players_v1(div)
        if not is_players_empty():
            return players

    return players


def _parse(
    html: str, players: list[list[tuple[str, str, str, str, Position]]]
) -> tuple[
    list[float], datetime.datetime, list[list[tuple[str, str, str, str, Position]]]
]:
    soup = BeautifulSoup(html, "lxml")
    return _extract_odds(soup), _extract_dt(soup), _extract_players(soup, players)


def create_afl_afl_game_model(
    team_names: list[str],
    players: list[list[tuple[str, str, str, str, Position]]],
    dt: datetime.datetime | None,
    venue_name: str,
    session: requests_cache.CachedSession,
    ladder: list[str],
    url: str | None,
    playwright: Playwright,
    version: str,
) -> GameModel:
    """Create a game model from AFL Tables."""
    odds: list[float] = []
    if url is not None and not pytest_is_running.is_running():
        ensure_install()
        browser = playwright.chromium.launch()
        context = browser.new_context()
        page = context.new_page()
        try:
            page.goto(url + "#line-ups", wait_until="networkidle")
        except:  # noqa: E722
            pass
        odds, dt, players = _parse(page.content(), players)
    if dt is None:
        raise ValueError("dt is null")

    venue_model = create_afl_afl_venue_model(
        venue_name=venue_name, session=session, dt=dt, version=VENUE_VERSION
    )
    teams = [
        create_afl_afl_team_model(
            team_name=x,
            players=players[count],
            session=session,
            dt=dt,
            ladder=ladder,
            odds=odds[count] if count < len(odds) else None,
            version=VERSION,
        )
        for count, x in enumerate(team_names)
    ]
    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=venue_model,
        teams=teams,
        end_dt=None,
        attendance=None,
        league=League.AFL,
        year=datetime.datetime.today().year,
        season_type=None,
        postponed=None,
        play_off=None,
        distance=None,
        dividends=[],
        pot=None,
        version=version,
    )
